from __future__ import annotations

"""
Base Types
"""

from pydantic import BaseModel as _BaseModel
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ..configs import AuthZeroSettings
    from ..flows import AZFlow, AZFlowSchema, AZManagementAPI
    from . import AZResourceSchema, AZResource

class BaseModel(_BaseModel):
    """
    Base Model
    """

    model_config = {'arbitrary_types_allowed': True}

    @property
    def settings(self) -> 'AuthZeroSettings':
        """
        Returns the settings
        """
        from ..utils.lazy import get_az_settings
        return get_az_settings()
    

    @classmethod
    def get_settings(cls) -> 'AuthZeroSettings':
        """
        Returns the settings
        """
        from ..utils.lazy import get_az_settings
        return get_az_settings()
    
    @classmethod
    def get_flow_schema(cls, name: str) -> 'AZFlowSchema':
        """
        Returns the AZFlowSchema
        """
        from ..utils.lazy import get_az_flow_schema
        return get_az_flow_schema(name)
    
    @classmethod
    def get_flow(cls, name: str, *args, **kwargs) -> 'AZFlow':
        """
        Returns the flow
        """
        from ..utils.lazy import get_az_flow
        return get_az_flow(name, *args, **kwargs)
    
    @classmethod
    def get_resource_schema(cls, name: str) -> 'AZResourceSchema':
        """
        Returns the resource schema
        """
        from ..utils.lazy import get_az_resource_schema
        return get_az_resource_schema(name)
    
    @classmethod
    def get_resource(cls, name: str, *args, **kwargs) -> 'AZResource':
        """
        Returns the resource
        """
        from ..utils.lazy import get_az_resource
        return get_az_resource(name, *args, **kwargs)
    
    @classmethod
    def get_management_api(cls) -> 'AZManagementAPI':
        """
        Returns the management api
        """
        from ..utils.lazy import get_az_mtg_api
        return get_az_mtg_api()
    
    


from __future__ import annotations

"""
Proxied Client
"""
import json
import asyncio
import contextlib
from lazyops.types.lazydict import LazyDict
from lazyops.utils.times import Timer
from lazyops.utils.logs import Logger
from lazyops.utils.helpers import create_timestamp, adeferred_task
from lazyops.utils.pooler import ThreadPooler
from proposalflow.utils.logs import logger, null_logger
from typing import Any, Dict, List, Optional, Type, TypeVar, Tuple, Union, Set, TYPE_CHECKING


if TYPE_CHECKING:
    from kvdb import KVDBSession, PersistentDict
    from lazyops.libs.logging import Logger
    from lazyops.libs.abcs.configs.base import AppSettings
    from lazyops.libs.abcs.types.api import BaseSchema
    from lazyops.libs.abcs.clients import ClientTypes
    from lazyops.utils.pooler import ThreadPooler, ThreadPool

class BaseProxyClient(LazyDict):
    """
    A Proxy Dictionary that lazily defers initialization of the 
    clients until they are called
    """
    name: Optional[str] = "proxy"
    kind: Optional[str] = "client"

    exclude_schema_attrs: Optional[bool] = True
    proxy_schema: Optional[Dict[str, str]] = None

    def __init__(self, **kwargs):
        """
        Handles the initialization of the proxy
        """
        self._dict = {}
        self.excluded_attrs = []
        self.excluded_attrs.extend(iter(dir(self)))
        if self.proxy_schema is None:
            self.proxy_schema = self.settings.ctx.get_client_registry_mapping()
        # if self.proxy_schema is not None:
        self.excluded_attrs.extend(self.proxy_schema.keys())
        self.excluded_attrs = list(set(self.excluded_attrs))
        self.logger.info(f"Initialized {self.name} Proxy Client: {self.excluded_attrs}")
        self.post_init(**kwargs)

    def post_init(self, **kwargs):
        """
        Post Initialization to be overwritten by the subclass
        """
        pass
    
    def get_or_init(self, name: str, default: Any = None) -> 'ClientTypes':
        """
        Gets the component
        """
        if name not in self._dict and name not in self.proxy_schema:
            raise ValueError(f"Invalid component {name} for {self.name}.{self.kind}")
        if name not in self._dict:
            self._dict[name] = self.settings.ctx.get_client(
                name = name,
            )
        return self._dict[name]


    @property
    def logger(self) -> 'Logger':
        """
        Gets the logger
        """
        return self.settings.logger
    
    @property
    def null_logger(self) -> 'Logger':
        """
        Gets the null logger
        """
        return self.settings.null_logger
    

    @property
    def autologger(self) -> 'Logger':
        """
        Automatic Logger that is enabled in devel mode
        """
        return self.settings.autologger

    
    @property
    def settings(self) -> 'AppSettings':
        """
        Gets the settings
        """
        if self._settings is None:
            from lazyops.libs.abcs.configs.lazy import get_module_settings
            self._settings = get_module_settings(self.__module__.__name__)
        return self._settings

    @property
    def kdb(self) -> 'KVDBSession':
        """
        Gets the keydb
        """
        if self._kdb is None:
            self._kdb = self.settings.ctx.get_kdb_session(self.name)
        return self._kdb
    
    @property
    def pooler(self) -> 'ThreadPool':
        """
        Gets the pooler
        """
        if self._pooler is None:
            from lazyops.libs.pooler import ThreadPooler
            self._pooler = ThreadPooler
        return self._pooler
    
    


"""
The PostHog Client

An Asyncronous PostHog Client that utilizes Intelligent Batching and Asynchronous Task Queues
to send events to PostHog in a more efficient and scalable manner.

This client is designed with the following features in mind:
- Intelligent Batching: The client automatically batches events based on the configured batch size and sends them in batches once the batch size is reached or the maximum batch interval has been reached.
- Asynchronous Task Queues: The client uses asynchronous task queues to handle the sending of events. This allows the client to send events in parallel and improves the overall performance of the client.
- Context Manager: The client provides a context manager that can be used to force sending of events immediately after exiting the context rather than waiting for the batching logic to complete.
- Non-Blocking: By utilizing asynchronous task queues and non-blocking requests, the client can handle a large number of events without blocking the main thread.
- Lazily Initializes: The client lazily initializes the PostHog client and only initializes it when the first event is added.


Usage:

# Initialize the PostHog Client
# you can capture events explicitly by using the client.capture method or by using the @capture decorator
# to defer until the function is called. If `POSTHOG_API_KEY` is not set, then the client will not be initialized

from lazyops.libs.posthog import PostHogClient, capture, PHCtxT

# This will lazily initialize the client since the client is not yet defined

@capture(event = 'lazy_event')
async def lazy_event(*args, x: Optional[int] = None, y: Optional[int] = None, ph_ctx: PHCtxT, **kwargs):
    
    ph_ctx['properties']['event_name'] = 'lazy_event'
    ph_ctx['properties']['random_id'] = create_unique_id()
    if x: ph_ctx['properties']['x'] = x
    if y: ph_ctx['properties']['y'] = y
    ph_ctx['distinct_id'] = f'xxx-{create_unique_id()}'
    return 'test'


client = PostHogClient(api_key = 'xxx', project_id = 'xxx')

# Explicitly Capture an Event with the Client since it has been initialized

@client.capture(event = 'test_event')
async def capture_event(*args, x: Optional[int] = None, y: Optional[int] = None, ph_ctx: PHCtxT, **kwargs):
    
    ph_ctx['properties']['event_name'] = 'test_event'
    ph_ctx['properties']['random_id'] = create_unique_id()
    if x: ph_ctx['properties']['x'] = x
    if y: ph_ctx['properties']['y'] = y
    ph_ctx['distinct_id'] = f'xxx-{create_unique_id()}'
    
    return 'test'


"""

from .config import PostHogSettings
from .client import PostHogClient
from .types import PostHogEndpoint, CaptureEvent, IdentifyEvent, EventQueue, EventT, PHCtxT
from .utils import get_posthog_client, get_posthog_settings, capture




"""
Niquest Based Client
"""

from __future__ import annotations

from lazyops.libs import lazyload


from gspread.client import Client as BaseClient
from gspread.http_client import HTTPClient as BaseHTTPClient
from typing import Optional, Union, Tuple

if lazyload.TYPE_CHECKING:
    import niquests
    import requests

    from google.auth.credentials import Credentials
    from google.auth.transport.requests import AuthorizedSession
    from niquests import Request
else:
    niquests = lazyload.LazyLoad("niquests")


class CredsAuth(niquests.Auth):
    """
    An instance of this class is used to authenticate requests.
    """

    def __init__(self, auth: Credentials) -> None:
        self.auth = auth

    def __call__(self, r: 'Request'):
        
        if not self.auth.valid:
            self.auth.refresh(r)


class HTTPClient(BaseHTTPClient):
    """An instance of this class communicates with Google API.

    :param Credentials auth: An instance of google.auth.Credentials used to authenticate requests
        created by either:

        * gspread.auth.oauth()
        * gspread.auth.oauth_from_dict()
        * gspread.auth.service_account()
        * gspread.auth.service_account_from_dict()

    :param Session session: (Optional) An OAuth2 credential object. Credential objects
        created by `google-auth <https://github.com/googleapis/google-auth-library-python>`_.

        You can pass you own Session object, simply pass ``auth=None`` and ``session=my_custom_session``.

    This class is not intended to be created manually.
    It will be created by the gspread.Client class.
    """

    def __init__(self, auth: 'Credentials', session: Optional['requests.Session'] = None) -> None:


        if session is not None:
            self.session = session
        else:
            from gspread.utils import convert_credentials
            self.auth: Credentials = convert_credentials(auth)
            self.session = AuthorizedSession(self.auth)

        self.timeout: Optional[Union[float, Tuple[float, float]]] = None



from __future__ import annotations

"""
System Monitoring Client
"""

import abc
import time
import atexit
import contextlib

from lazyops.libs import lazyload
from pydantic import BaseModel
from pydantic.types import ByteSize
from lazyops.utils import logger, Timer
from typing import Optional, List, Dict, Any, Union, Type, Set, TYPE_CHECKING

if lazyload.TYPE_CHECKING:
    import psutil
else:
    psutil = lazyload.LazyLoad("psutil")


class SysMonitorData(BaseModel):
    """
    The System Monitor Data
    """
    cpu: Dict[str, float]
    memory: Dict[str, float]
    disk: Dict[str, float]
    network: Dict[str, float]
    process: Dict[str, float]
    system: Dict[str, float]

class SysMonitorClient(abc.ABC):
    """
    The System Monitor Client
    """

    def __init__(
        self,
        interval: Optional[float] = None, # If interval is set, background tasks will be created
        **kwargs
    ):
        """
        Initializes the System Monitor Client
        """
        self.interval = interval
        self.timer = Timer()
        self.states: Dict[int, SysMonitorData] = {}
    
    def capture(self) -> SysMonitorData:
        """
        Captures the system monitor data
        """
        return SysMonitorData(
            cpu = self.get_cpu(),
            memory = self.get_memory(),
            disk = self.get_disk(),
            network = self.get_network(),
            process = self.get_process(),
            system = self.get_system(),
        )
    
    def get_cpu(self) -> Dict[str, float]:
        """
        Returns the CPU Usage
        """
        return psutil.cpu_percent(interval = self.interval)
    
    def get_memory(self) -> Dict[str, float]:
        """
        Returns the Memory Usage
        """
        mem = psutil.virtual_memory()
        return {
            'total': mem.total,
            'available': mem.available,
            'percent': mem.percent,
            'used': mem.used,
            'free': mem.free,
            'active': mem.active,
            'inactive': mem.inactive,
        }


from __future__ import annotations

"""
Persistent Dict Addons
"""

from pydantic import BaseModel, Field
from typing import List, Optional, Union, Tuple, Dict, Iterator, Any, Type, Set, Iterable, TYPE_CHECKING
from .debug import get_autologger

autologger = get_autologger('addons')

class NumericValuesContainer(BaseModel):
    """
    Container for numeric values
    """
    values: Optional[List[float]] = Field(default_factory=list, description="The Values of the container")

    @property
    def average(self) -> float:
        """
        Returns the average
        """
        return sum(self.values) / len(self.values) if self.values else 0
    
    @property
    def median(self) -> float:
        """
        Returns the median
        """
        return sorted(self.values)[len(self.values) // 2] if self.values else 0
    
    @property
    def count(self) -> int:
        """
        Returns the count
        """
        return len(self.values) if self.values else 0
    
    @property
    def min(self) -> float:
        """
        Returns the min
        """
        return min(self.values) if self.values else 0
    
    @property
    def max(self) -> float:
        """
        Returns the max
        """
        return max(self.values) if self.values else 0
    
    @property
    def total(self) -> float:
        """
        Returns the total
        """
        return sum(self.values) if self.values else 0

    @property
    def p90(self) -> float:
        """
        Returns the p90
        """
        return self.percentile(90)
    
    @property
    def p95(self) -> float:
        """
        Returns the p95
        """
        return self.percentile(95)
    
    def percentile(self, percentile: float) -> float:
        """
        Returns the percentile
        """
        return sorted(self.values)[int(len(self.values) * (percentile / 100))] if self.values else 0

    def __len__(self) -> int:
        """
        Returns the length
        """
        return len(self.values) if self.values else 0

    def __iadd__(self, other: Union[float, int]) -> 'NumericValuesContainer':
        """
        Adds the other to the values
        """
        self.values.append(other)
        return self
    
    def __isub__(self, other: Union[float, int]) -> 'NumericValuesContainer':
        """
        Subtracts the other from the values
        """
        self.values.append(other * -1)
        return self
    
    def __add__(self, other: Union[float, int]) -> float:
        """
        Adds the other to the values
        """
        return self.total + other
    
    def __sub__(self, other: Union[float, int]) -> float:
        """
        Subtracts the other from the values
        """
        return self.total - other
    
    def __mul__(self, other: Union[float, int]) -> float:
        """
        Multiplies the other to the values
        """
        return self.total * other
    
    def __truediv__(self, other: Union[float, int]) -> float:
        """
        Divides the other to the values
        """
        return self.total / other
    
    def __floordiv__(self, other: Union[float, int]) -> float:
        """
        Divides the other to the values
        """
        return int(self.total / other)
    
    def __mod__(self, other: Union[float, int]) -> float:
        """
        Divides the other to the values
        """
        return (self.total % other)
    
    def __divmod__(self, other: Union[float, int]) -> Tuple[float, float]:
        """
        Divides the other to the values
        """
        return divmod(self.total, other)
    
    def pop(self, index: int) -> float:
        """
        Pops the item at the index
        """
        return self.values.pop(index)
    
    def append(self, value: float) -> None:
        """
        Appends the value
        """
        self.values.append(value)

    def extend(self, values: List[float]) -> None:
        """
        Extends the values
        """
        self.values.extend(values)

    def __getitem__(self, index: int) -> float:
        """
        Gets the item at the index
        """
        return self.values[index]
    
    def __setitem__(self, index: int, value: float) -> None:
        """
        Sets the item at the index
        """
        self.values[index] = value

    def __delitem__(self, index: int) -> None:
        """
        Deletes the item at the index
        """
        del self.values[index]

    def __iter__(self) -> Iterator[float]:
        """
        Iterates over the values
        """
        return iter(self.values)
    
    def __contains__(self, value: float) -> bool:
        """
        Checks if the value is in the values
        """
        return value in self.values

    def __reversed__(self) -> Iterator[float]:
        """
        Reverses the values
        """
        return reversed(self.values)
    
    def __bool__(self) -> bool:
        """
        Checks if the values are not empty
        """
        return bool(self.values)
    
    def __eq__(self, other: float) -> bool:
        """
        Checks if the values are equal
        """
        return self.total == other
    
    def __ne__(self, other: float) -> bool:
        """
        Checks if the values are not equal
        """
        return self.total != other
    
    def __lt__(self, other: float) -> bool:
        """
        Checks if the values are less than
        """
        return self.total < other
    
    def __le__(self, other: float) -> bool:
        """
        Checks if the values are less than or equal to
        """
        return self.total <= other
    
    def __gt__(self, other: float) -> bool:
        """
        Checks if the values are greater than
        """
        return self.total > other
    
    def __ge__(self, other: float) -> bool:
        """
        Checks if the values are greater than or equal to
        """
        return self.total >= other
    
    def __hash__(self) -> int:
        """
        Returns the hash of the values
        """
        return hash(self.values)
    
    def __repr__(self) -> str:
        """
        Returns the representation of the values
        """
        return f'<{self.__class__.__name__}>(total={self.total}, average={self.average}, count={self.count})'

    def __str__(self) -> str:
        """
        Returns the string representation of the values
        """
        return f'{self.total}, {self.average}, {self.count}'
    
    @property
    def data_values(self) -> Dict[str, Union[float, int]]:
        """
        Returns the data view
        """
        return {
            'total': self.total,
            'average': self.average,
            'median': self.median,
            'count': self.count,
        }

class MonetaryMetric(NumericValuesContainer):
    """
    A container for monetary values
    """
    name: Optional[str] = 'costs'
    unit: Optional[str] = 'USD'
    symbol: Optional[str] = '$'

    def pretty(self, value: float) -> str:
        """
        Returns the pretty representation of the value
        """
        return f'{self.symbol}{value:.2f}'
    
    @property
    def total_s(self) -> str:
        """
        Returns the total in s
        """
        return self.pretty(self.total)
    
    @property
    def average_s(self) -> str:
        """
        Returns the average in s
        """
        return self.pretty(self.average)
    
    @property
    def median_s(self) -> str:
        """
        Returns the median in s
        """
        return self.pretty(self.median)
    
    def __repr__(self) -> str:
        """
        Returns the representation of the values
        """
        d = {
            self.name: {
                'total': self.total_s,
                'average': self.average_s,
                'median': self.median_s,
                'count': self.count,
            }
        }
        return str(d)
    
    def __str__(self) -> str:
        """
        Returns the string representation of the values
        """
        return f'{self.total_s}, {self.average_s}, {self.count}'



class DurationMetric(NumericValuesContainer):
    """
    A container for duration metrics
    """

    name: Optional[str] = 'duration'


    @classmethod
    def dformat_duration(
        cls,
        duration: float,
        pretty: bool = True,
        short: int = 0,
        include_ms: bool = False,
        as_int: bool = False,
    ) -> Dict[str, Union[float, int]]:    # sourcery skip: low-code-quality
        """
        Formats a duration (secs) into a dict
        """
        if not pretty:
            unit = 'secs' if short else 'seconds'
            value = int(duration) if as_int else duration
            return {unit: value}
        data = {}
        if duration >= 86400:
            unit = 'd' if short > 1 else 'day'
            days = (duration // 86400)
            if short < 2 and days > 1: unit += 's'
            duration -= days * 86400
            data[unit] = int(days) if as_int else days
        if duration >= 3600:
            unit = 'hr' if short else 'hour'
            if short > 1: unit = unit[0]
            hours = (duration // 3600)
            if short < 2 and hours > 1: unit += 's'
            duration -= hours * 3600
            data[unit] = int(hours) if as_int else hours
        if duration >= 60:
            unit = 'min' if short else 'minute'
            if short > 1: unit = unit[0]
            minutes = (duration // 60)
            if short < 2 and minutes > 1: unit += 's'
            duration -= minutes * 60
            data[unit] = int(minutes) if as_int else minutes
        if duration >= 1:
            unit = 'sec' if short else 'second'
            if short > 1: unit = unit[0]
            if short < 2 and duration > 1: unit += 's'
            if include_ms:
                seconds = int(duration)
                duration -= seconds
                data[unit] = seconds
            elif short > 1:
                data[unit] = int(duration) if as_int else duration
            else:
                data[unit] = float(f'{duration:.2f}')
        if include_ms and duration > 0:
            unit = 'ms' if short else 'millisecond'
            milliseconds = int(duration * 1000)
            data[unit] = milliseconds
        return data
    
    @classmethod
    def pformat_duration(
        cls,
        duration: float,
        pretty: bool = True,
        short: int = 0,
        include_ms: bool = False,
    ) -> str:  # sourcery skip: low-code-quality
        """
        Formats a duration (secs) into a string

        535003.0 -> 5 days, 5 hours, 50 minutes, 3 seconds
        3593.0 -> 59 minutes, 53 seconds
        """
        data = cls.dformat_duration(
            duration = duration,
            pretty = pretty,
            short = short,
            include_ms = include_ms,
            as_int = True
        )
        if not data: return '0 secs'
        sep = '' if short > 1 else ' '
        if short > 2: return ''.join([f'{v}{sep}{k}' for k, v in data.items()])
        return ' '.join([f'{v}{sep}{k}' for k, v in data.items()]) if short else ', '.join([f'{v}{sep}{k}' for k, v in data.items()])

    def pretty(
        self,
        short: int = 0,
        include_ms: bool = False,
    ) -> str:
        """
        Returns the pretty representation of the values
        """
        return self.pformat_duration(self.total, short = short, include_ms = include_ms)

    def __repr__(self) -> str:
        """
        Returns the representation of the values
        """
        name = self.name or self.__class__.__name__
        return f'<{name}>(total: {self.pformat_duration(self.total)}, average: {self.average:.2f}/sec)'

    def __str__(self) -> str:
        """
        Returns the string representation of the values
        """
        return f'{self.pformat_duration(self.total)}'
    
    @property
    def total_s(self) -> str:
        """
        Returns the total in seconds
        """
        return self.pformat_duration(self.total, short = 1, include_ms = False)
    
    @property
    def average_s(self) -> str:
        """
        Returns the average in seconds
        """
        return self.pformat_duration(self.average, short = 1, include_ms = False)
    
    @property
    def median_s(self) -> str:
        """
        Returns the median in seconds
        """
        return self.pformat_duration(self.median, short = 1, include_ms = False)


class CountMetric(BaseModel):
    """
    A container for count metrics
    This is a generic metric that can be used to track individual counts of keys

    {
        'website.com': 1,
        'www.website.com': 2,
    }
    """

    name: Optional[str] = 'count'
    data: Dict[str, Union[int, float]] = Field(default_factory = dict, description = 'The count values')

    def incr(self, key: str, value: Union[int, float] = 1) -> None:
        """
        Increments the value for the given key
        """
        if key not in self.data: self.data[key] = value
        else: self.data[key] += value

    def decr(self, key: str, value: Union[int, float] = 1) -> None:
        """
        Decrements the value for the given key
        """
        if key not in self.data: self.data[key] = value
        else: self.data[key] -= value

    def reset(self, key: Optional[str] = None) -> None:
        """
        Resets the values
        """
        if key is not None: self.data[key] = 0
        else: self.data.clear()

    def add(self, key: str) -> None:
        """
        Adds the value to the count
        """
        return self.incr(key)

    def sub(self, key: str) -> None:
        """
        Subtracts the value from the count
        """
        return self.decr(key)
    
    def append(self, key: str) -> None:
        """
        Appends the value to the count
        """
        return self.incr(key, value = 1)
    
    def extend(self, *keys: str) -> None:
        """
        Extends the count
        """
        for key in keys: self.incr(key, value = 1)
    
    def __getitem__(self, key: str) -> Union[int, float]:
        """
        Gets the value for the given key
        """
        return self.data[key]
    
    def __setitem__(self, key: str, value: Union[int, float]) -> None:
        """
        Sets the value for the given key
        """
        self.data[key] = value

    def __delitem__(self, key: str) -> None:
        """
        Deletes the value for the given key
        """
        del self.data[key]

    def __contains__(self, key: str) -> bool:
        """
        Checks if the key is in the values
        """
        return key in self.data
    
    def __len__(self) -> int:
        """
        Gets the length of the values
        """
        return len(self.data)
    
    def items(self, sort: Optional[bool] = None): #  -> List[Tuple[str, Union[int, float, CountMetric]]]:
        """
        Gets the items of the values
        """
        if sort: return sorted(self.data.items(), key = lambda x: x[1], reverse = True)
        return list(self.data.items())
    
    def keys(self, sort: Optional[bool] = None) -> List[str]:
        """
        Gets the keys of the values
        """
        if sort: return sorted(self.data.keys(), key = lambda x: x, reverse = True)
        return list(self.data.keys())
    
    def values(self, sort: Optional[bool] = None) -> List[Union[int, float]]:
        """
        Gets the values of the values
        """
        if sort: return sorted(self.data.values(), key = lambda x: x, reverse = True)
        return list(self.data.values())
    
    def __iter__(self) -> Iterable[str]:
        """
        Iterates over the values
        """
        return iter(self.data.keys())
    
    def __iadd__(self, key: Union[str, List[str]]) -> 'CountMetric':
        """
        Adds a key to the values
        """
        if not isinstance(key, list): key = [key]
        for k in key:
            self.incr(k)
        return self
    
    def __isub__(self, key: Union[str, List[str]]) -> 'CountMetric':
        """
        Subtracts a key from the values
        """
        if not isinstance(key, list): key = [key]
        for k in key:
            self.decr(k)
        return self

    @property
    def count(self) -> int:
        """
        Gets the count of the values
        """
        return len(self.data)
    
    @property
    def total(self) -> Union[int, float]:
        """
        Gets the total value
        """
        return sum(self.data.values())
    
    @property
    def average(self) -> Union[int, float]:
        """
        Gets the average value
        """
        return self.total / self.count
    
    def top_n(self, n: int, sort: Optional[bool] = None) -> List[Tuple[str, Union[int, float]]]:
        """
        Gets the top n values
        """
        if sort: return sorted(self.data.items(), key = lambda x: x[1], reverse = True)[:n]
        return list(self.data.items())[:n]
    
    def top_n_keys(self, n: int, sort: Optional[bool] = None) -> List[str]:
        """
        Gets the top n keys
        """
        if sort: return sorted(self.data.keys(), key = lambda x: x, reverse = True)[:n]
        return list(self.data.keys())[:n]
    
    def top_n_values(self, n: int, sort: Optional[bool] = None) -> List[Union[int, float]]:
        """
        Gets the top n values
        """
        if sort: return sorted(self.data.values(), key = lambda x: x, reverse = True)[:n]
        return list(self.data.values())[:n]
    
    def top_n_items(self, n: int, sort: Optional[bool] = None) -> Dict[str, Union[int, float]]:
        """
        Gets the top n items
        """
        if sort: return dict(sorted(self.data.items(), key = lambda x: x[1], reverse = True)[:n])
        return dict(list(self.data.items())[:n])
    
    def __repr__(self) -> str:
        """
        Representation of the object
        """
        return f'{dict(self.items())}'

    @property
    def key_list(self) -> List[str]:
        """
        Returns the list of keys
        """
        return list(self.data.keys())


MetricModel = Union[CountMetric, DurationMetric, MonetaryMetric, NumericValuesContainer, BaseModel]


class BaseNestedMetric(BaseModel):
    """
    The base class for nested metrics

    - This is a base class for metrics that contain nested metrics
    """
    name: Optional[str] = None
    data: Dict[str, MetricModel] = Field(default_factory = dict, description = 'The nested metric values')

    @property
    def metric_class(self) -> Type[MetricModel]:
        """
        Returns the metric class
        """
        raise NotImplementedError

    def items(self, sort: Optional[bool] = None):
        """
        Returns the dict_items view of the data
        """
        return {k: dict(v.items(sort = sort)) for k, v in self.data.items()}.items()

    def __getitem__(self, key: str) -> MetricModel:
        """
        Gets the value for the given key
        """
        if key not in self.data: self.data[key] = self.metric_class(name = key)
        return self.data[key]
    
    def __setitem__(self, key: str, value: MetricModel):
        """
        Sets the value for the given key
        """
        self.data[key] = value

    def __getattr__(self, name: str) -> MetricModel:
        """
        Gets the value for the given key
        """
        if name not in self.data: self.data[name] = self.metric_class(name = name)
        return self.data[name]
    
    def __setattr__(self, name: str, value: MetricModel) -> None:
        """
        Sets the value for the given key
        """
        self.data[name] = value
    
    def __repr__(self) -> str:
        """
        Representation of the object
        """
        return f'{dict(self.items())}'
        
    def __str__(self) -> str:
        """
        Representation of the object
        """
        return self.__repr__()

    def __contains__(self, key: str) -> bool:
        """
        Checks if the key is in the data
        """
        return key in self.data
    
    
    if TYPE_CHECKING:
        def __getattribute__(self, name: str) -> MetricModel:
            """
            Gets the value for the given key
            """
            ...


class NestedCountMetric(BaseNestedMetric):
    """
    Nested Count Metric Container

    {
        'website.com': {
            '2022-01-01': 1,
            '2022-01-02': 2,
        },
        'www.website.com': {
            '2022-01-01': 3,
            '2022-01-02': 4,
        },
    }
    """

    name: Optional[str] = 'nested_count'
    data: Dict[str, CountMetric] = Field(default_factory = dict, description = 'The nested count values')


    @property
    def metric_class(self) -> Type[CountMetric]:
        """
        Returns the metric class
        """
        return CountMetric

    def items(self, sort: Optional[bool] = None):
        """
        Returns the dict_items view of the data
        """
        return {k: dict(v.items(sort = sort)) for k, v in self.data.items()}.items()

    if TYPE_CHECKING:
        def __getitem__(self, key: str) -> CountMetric:
            """
            Gets the value for the given key
            """
            ...
        def __getattr__(self, name: str) -> CountMetric:
            """
            Gets the value for the given key
            """
            ...


class NestedDurationMetric(BaseNestedMetric):
    """
    Nested Duration Metric Container

    {
        'operation_a': 240.0,
        'operation_b': 120.0,
    }
    """
    name: Optional[str] = 'nested_duration'
    data: Dict[str, DurationMetric] = Field(default_factory = dict, description='The nested duration metric data')

    @property
    def metric_class(self) -> Type[DurationMetric]:
        """
        Returns the metric class
        """
        return DurationMetric

    @property
    def data_values(self) -> Dict[str, float]:
        """
        Returns the data values
        """
        return {k: v.total for k, v in self.data.items()}

    def items(self, sort: Optional[bool] = None):
        """
        Returns the dict_items view of the data
        """
        if sort: return dict(sorted(self.data.items(), key = lambda x: x[1].total, reverse = sort))
        return self.data_values.items()

    if TYPE_CHECKING:
        def __getitem__(self, key: str) -> DurationMetric:
            """
            Gets the value for the given key
            """
            ...
        def __getattr__(self, name: str) -> DurationMetric:
            """
            Gets the value for the given key
            """
            ...

    
class NestedMonetaryMetric(BaseNestedMetric):
    """
    Nested Monetary Metric Container
    {
        'operation_a': {
            'total': 100.0,
            'average': 50.0,
            'median': 25.0,
            'count': 10,
        },
        'operation_b': {
            'total': 200.0,
            'average': 100.0,
            'median': 50.0,
            'count': 10,
        }
    }
    """
    name: Optional[str] = 'nested_monetary'
    data: Dict[str, MonetaryMetric] = Field(default_factory = dict, description='The nested monetary metric data')

    @property
    def metric_class(self) -> Type[MonetaryMetric]:
        """
        Returns the metric class
        """
        return MonetaryMetric

    @property
    def data_values(self) -> Dict[str, float]:
        """
        Returns the data values

        {
            'operation_a': {
                'total': 100.0,
                'average': 50.0,
                'median': 25.0,
                'count': 10,
            },
            'operation_b': {
                'total': 200.0,
                'average': 100.0,
                'median': 50.0,
                'count': 10,
            }
        }
        """
        return {k: v.data_values for k, v in self.data.items()}


    def items(self, **kwargs):
        """
        Returns the dict_items view of the data
        """
        return self.data_values.items()

    if TYPE_CHECKING:
        def __getitem__(self, key: str) -> MonetaryMetric:
            """
            Gets the value for the given key
            """
            ...
        def __getattr__(self, name: str) -> MonetaryMetric:
            """
            Gets the value for the given key
            """
            ...


class NestedCountMetricV1(BaseModel):
    """
    Nested Count Metric Container

    {
        'website.com': {
            '2022-01-01': 1,
            '2022-01-02': 2,
        },
        'www.website.com': {
            '2022-01-01': 3,
            '2022-01-02': 4,
        },
    }
    """

    name: Optional[str] = 'nested_count'
    data: Dict[str, CountMetric] = Field(default_factory = dict, description = 'The nested count values')

    def items(self, sort: Optional[bool] = None):
        """
        Returns the dict_items view of the data
        """
        return {k: dict(v.items(sort = sort)) for k, v in self.data.items()}.items()

    def __getitem__(self, key: str) -> CountMetric:
        """
        Gets the value for the given key
        """
        if key not in self.data: self.data[key] = CountMetric(name = key)
        return self.data[key]
    
    def __setitem__(self, key: str, value: CountMetric):
        """
        Sets the value for the given key
        """
        # autologger.info(f'Setting {key} to {value}', prefix = self.name, colored = True)
        self.data[key] = value

    def __getattr__(self, name: str) -> CountMetric:
        """
        Gets the value for the given key
        """
        if name not in self.data: self.data[name] = CountMetric(name = name)
        return self.data[name]
    
    def __setattr__(self, name: str, value: CountMetric) -> None:
        """
        Sets the value for the given key
        """
        self.data[name] = value
    
    def __repr__(self) -> str:
        """
        Representation of the object
        """
        return f'{dict(self.items())}'
        
    def __str__(self) -> str:
        """
        Representation of the object
        """
        return self.__repr__()
    

class NestedDurationMetricV1(BaseModel):
    """
    Nested Duration Metric Container

    {
        'operation_a': 240.0,
        'operation_b': 120.0,
    }
    """
    name: Optional[str] = 'nested_duration'
    data: Dict[str, DurationMetric] = Field(default_factory = dict, description='The nested duration metric data')

    @property
    def data_values(self) -> Dict[str, float]:
        """
        Returns the data values
        """
        return {k: v.total for k, v in self.data.items()}

    def items(self, sort: Optional[bool] = None):
        """
        Returns the dict_items view of the data
        """
        if sort: return dict(sorted(self.data.items(), key = lambda x: x[1].total, reverse = sort))
        return self.data_values.items()

    def __getitem__(self, key: str) -> DurationMetric:
        """
        Gets the value for the given key
        """
        if key not in self.data: self.data[key] = DurationMetric(name = key)
        return self.data[key]
    
    def __setitem__(self, key: str, value: DurationMetric):
        """
        Sets the value for the given key
        """
        self.data[key] = value

    def __getattr__(self, name: str) -> DurationMetric:
        """
        Gets the value for the given key
        """
        if name not in self.data: self.data[name] = DurationMetric(name = name)
        return self.data[name]
    
    def __setattr__(self, name: str, value: DurationMetric) -> None:
        """
        Sets the value for the given key
        """
        self.data[name] = value
    
    def __repr__(self) -> str:
        """
        Representation of the object
        """
        return f'{dict(self.items())}'
        
    def __str__(self) -> str:
        """
        Representation of the object
        """
        return self.__repr__()
    

MetricT = Union[
    CountMetric, 
    DurationMetric, 
    MonetaryMetric, 
    NestedDurationMetric, 
    NestedCountMetric, 
    NestedMonetaryMetric,
    NumericValuesContainer, 
]


from __future__ import annotations

import os
import abc
import signal
import contextlib
import multiprocessing
import pathlib
import asyncio

from typing import Optional, List, TypeVar, Callable, Dict, Any, overload, Type, Union, TYPE_CHECKING
from lazyops.utils.lazy import lazy_import
from lazyops.libs.proxyobj import ProxyObject, proxied
from lazyops.imports._psutil import _psutil_available

if _psutil_available:
    import psutil

if TYPE_CHECKING:
    from lazyops.utils.logs import Logger
    from ..configs.base import AppSettings
    from lazyops.types.models import BaseSettings
    from lazyops.libs.fastapi_utils.types.state import AppState

    from kvdb import TaskFunction, CronJob, TaskWorker, TaskQueue

    # with contextlib.suppress(ImportError):
    #     from kvdb import TaskWorker, TaskQueue
        # from aiokeydb.types.task_queue import TaskQueue
        # from aiokeydb.types.worker import Worker


# SettingsT = TypeVar('SettingsT', bound='AppSettings')


# class GlobalContextClass(abc.ABC):

@proxied
class GlobalContext(abc.ABC):
    """
    Global Context for FastAPI 
    """


    workers: Dict[str, Dict[str, Union[multiprocessing.Process, 'TaskWorker']]] = {}
    # queues: Dict[str, Dict[str, 'TaskQueue']] = {}
    server_processes: List['multiprocessing.Process'] = []
    # worker_processes: Dict[str, Dict[str, List['multiprocessing.Process']]] = {}

    on_close_funcs: List[Callable] = []

    settings_func: Optional[Union[Callable, str]] = None # type: ignore
    settings_config_func: Optional[Union[Callable, str]] = None # type: ignore
    get_worker_func: Optional[Union[Callable, str, List[str]]] = None # type: ignore
    # get_num_worker_func: Optional[Union[Callable, str]] = None # type: ignore
    # get_worker_names_func: Optional[Union[Callable, str]] = None
    
    
    def __init__(self, **kwargs):
        """
        Creates the global context
        """
        self._settings: Optional['AppSettings'] = None
        self._state: Optional['AppState'] = None
        self._logger: Optional['Logger'] = None

    def configure(
        self,
        # queue_types: Optional[List[str]] = None,
        settings_func: Optional[Union[Callable, str]] = None, # type: ignore
        settings_config_func: Optional[Union[Callable, str]] = None, # type: ignore
        get_worker_func: Optional[Union[Callable, str, List[str]]] = None, # type: ignore
        # get_num_worker_func: Optional[Union[Callable, str]] = None, # type: ignore
        # get_worker_names_func: Optional[Union[Callable, str]] = None,
    ):
        """
        Configures the global context
        """
        # if queue_types is not None:
        #     for kind in queue_types:
        #         self.add_queue_type(kind)
        
        if settings_func is not None:
            if isinstance(settings_func, str):
                settings_func = lazy_import(settings_func)
            self.settings_func = settings_func
        
        if settings_config_func is not None:
            self.settings_config_func = settings_config_func
        
        if get_worker_func is not None:
            self.get_worker_func = get_worker_func
        
        # if get_num_worker_func is not None:
        #     self.get_num_worker_func = get_num_worker_func
        
        # if get_worker_names_func is not None:
        #     self.get_worker_names_func = get_worker_names_func
            

    @property
    def settings(self) -> 'AppSettings':
        """
        Returns the settings
        """
        if self._settings is None:
            if self.settings_func is None:
                raise RuntimeError("Settings not initialized. Set `GlobalContext.settings_func` to a function that returns the settings.")
            self._settings = self.settings_func()
        return self._settings
    
    @property
    def state(self) -> 'AppState':
        """
        Returns the state
        """
        if self._state is None:
            from lazyops.libs.fastapi_utils.types.state import AppState
            self._state = AppState()
            self._state.bind_settings(self.settings)
        return self._state
    
    @property
    def is_leader_process(self) -> bool:
        """
        Returns if this is the leader process
        """
        return self.state.is_leader_process or multiprocessing.parent_process() is None
    
    @property
    def is_primary_server_process(self) -> bool:
        """
        Returns if this is the primary server process
        """
        return self.state.is_primary_server_process or multiprocessing.parent_process() is None
    

    def get_settings_func(self, func: Union[Callable, str]) -> Callable:
        """
        Returns the settings func
        """
        if isinstance(func, str):
            func = getattr(self.settings, func)
        return func
    

    @property
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        if self._logger is None:
            from lazyops.utils.logs import logger as _logger
            self._logger = _logger
        return self._logger

    def get_child_process_ids(self, name: Optional[str] = None, kind: Optional[str] = None, process_id: Optional[int] = None, first: Optional[bool] = True) -> List[int]:
        """
        Returns the child process ids
        """
        if not _psutil_available:
            self.logger.warning("psutil not available. Cannot get child process ids")
            return []
        if not name and not process_id:
            raise ValueError("Must provide either name or process_id")
        if process_id:
            proc = psutil.Process(process_id)
            return [child.pid for child in proc.children(recursive=True)]
        if name in {'server', 'app'} and kind is None:
            procs = []
            for proc in self.server_processes:
                if proc._closed: continue
                parent = psutil.Process(proc.pid)
                procs.extend(child.pid for child in parent.children(recursive=True))
                if first: break
            return procs
        
        procs = []
        if name == 'worker':
            for worker_name, values in self.workers.items():
                process = values.get('process')
                if process is None or process._closed: continue
                parent = psutil.Process(process.pid)
                procs.extend(child.pid for child in parent.children(recursive=True))
                if first: break
        return procs
    
    @overload
    def start_task_workers(
        self,
        num_workers: Optional[int] = 1,
        start_index: Optional[int] = None,

        worker_name: Optional[str] = 'global',
        worker_name_sep: Optional[str] = '-',

        worker_imports: Optional[Union[str, List[str]]] = None,
        worker_cls: Optional[str] = None,
        worker_class: Optional[Type['TaskWorker']] = None,
        worker_config: Optional[Dict[str, Any]] = None,
        worker_queues: Optional[Union[str, List[str]]] = 'all',
        
        worker_functions: Optional[List['TaskFunction']] = None,
        worker_cron_jobs: Optional[List['CronJob']] = None,
        worker_startup: Optional[Union[List[Callable], Callable,]] = None,
        worker_shutdown: Optional[Union[List[Callable], Callable,]] = None,
        worker_before_process: Optional[Callable] = None,
        worker_after_process: Optional[Callable] = None,
        worker_attributes: Optional[Dict[str, Any]] = None,

        queue_names: Optional[Union[str, List[str]]] = None,
        queue_config: Optional[Dict[str, Any]] = None,
        queue_class: Optional[Type['TaskQueue']] = None,

        max_concurrency: Optional[int] = None,
        max_broadcast_concurrency: Optional[int] = None,

        debug_enabled: Optional[bool] = False,
        disable_env_name: Optional[bool] = False,
        method: Optional[str] = 'mp',
        use_new_event_loop: Optional[bool] = None,
        disable_worker_start: Optional[bool] = False,
        terminate_timeout: Optional[float] = 5.0,
        **kwargs,
    ):
        ...


    def start_task_workers(
        self,
        worker_name: Optional[str] = None,
        worker_imports: Optional[Union[str, List[str]]] = None,
        num_workers: Optional[int] = 1,
        worker_queues: Optional[Union[str, List[str]]] = 'all',
        debug_enabled: Optional[bool] = None,
        verbose: Optional[bool] = True,
        **kwargs,
    ):
        """
        Starts the task workers
        """
        
        if worker_name is None:
            worker_name = self.settings.module_name
        if worker_imports is None:
            worker_imports = self.get_worker_func
            if not isinstance(worker_imports, list):
                worker_imports = [worker_imports]
        if debug_enabled is None:
            debug_enabled = self.settings.debug_enabled or self.settings.is_development_env
        
        from kvdb import tasks
        workers = tasks.start_task_workers(
            worker_name = worker_name,
            worker_imports = worker_imports,
            num_workers = num_workers,
            worker_queues = worker_queues,
            debug_enabled = debug_enabled,
            **kwargs,
        )
        if verbose:
            self.logger.info(f"Started workers: {workers}")
        self.workers.update(workers)

    def stop_task_workers(
        self,
        worker_names: Optional[List[str]] = None,
        timeout: Optional[float] = 5.0,
        verbose: bool = True,
        **kwargs,
    ):
        """
        Stops the task workers
        """
        if worker_names is None:
            worker_names = list(self.workers.keys())
        from kvdb.tasks.spawn import exit_task_workers
        exit_task_workers(worker_names = worker_names, timeout = timeout)
        for worker_name in worker_names:
            if verbose: self.logger.info(f"Stopped worker: {worker_name}")
            _ = self.workers.pop(worker_name, None)

    
    def start_server_process(self, cmd: str, verbose: bool = True):
        """
        Starts the server process
        """
        if verbose: self.logger.info(f"Starting Server Process: {cmd}")
        context = multiprocessing.get_context('spawn')
        p = context.Process(target = os.system, args = (cmd,))
        p.start()
        self.state.add_leader_process_id(p.pid, 'server')
        self.server_processes.append(p)
        return p
    
    def add_server_process(self, name: str, cmd: str, verbose: bool = True):
        """
        Adds the server process
        """
        if verbose: self.logger.info(f"Adding Server Process: {cmd}", prefix = name)
        context = multiprocessing.get_context('spawn')
        p = context.Process(target = os.system, args = (cmd,))
        p.start()
        self.server_processes.append(p)
        return p

    def stop_server_processes(self, verbose: bool = True, timeout: float = 5.0):
        """
        Stops the server processes
        """
        curr_proc, n_procs = 0, len(self.server_processes)
        kind = 'Server'
        while self.server_processes:
            proc = self.server_processes.pop()
            if proc._closed: continue
            log_name = f'`|g|app-{curr_proc}|e|`'
            if verbose: 
                self.logger.info(f"- [|y|{kind:7s}|e|] Stopping: [ {curr_proc + 1}/{n_procs} ] {log_name:20s} (Process ID: {proc.pid})", colored = True)
            proc.join(timeout)
            proc.terminate()
            try:
                proc.close()
            except Exception as e:
                if verbose: self.logger.info(f"|y|[{kind:7s}]|e| Failed Stop: [ {curr_proc + 1}/{n_procs} ] {log_name:20s} (Error: |r|{e}|e|)", colored = True)
                try:
                    signal.pthread_kill(proc.ident, signal.SIGKILL)
                    proc.join(timeout)
                    proc.terminate()
                except Exception as e:
                    if verbose: self.logger.info(f"|r|[{kind:7s}]|e| Failed Kill: [ {curr_proc + 1}/{n_procs} ] {log_name:20s} (Error: |r|{e}|e|)", colored = True)
                with contextlib.suppress(Exception):
                    proc.kill()
                    proc.close()
            curr_proc += 1
        # if verbose: self.logger.info(f"Stopped all {curr_proc} server processes")
    
    def end_all_processes(self, verbose: bool = True, timeout: float = 5.0):
        """
        Terminates all processes
        """
        # for kind, names in self.worker_processes.items():
        #     for name in names:
        #         self.stop_worker_processes(name = name, verbose = verbose, timeout = timeout, kind = kind)
        self.stop_task_workers(timeout = timeout, verbose = verbose)
        self.stop_server_processes(verbose = verbose, timeout = timeout)
        # self.logger.info("Terminated all processes")

    def register_on_close(self, func: Callable, *args, **kwargs):
        """
        Registers a function to be called on close
        """
        import functools
        _func = functools.partial(func, *args, **kwargs)
        self.on_close_funcs.append(_func)
        self.logger.info(f"Registered function {func.__name__} to be called on close")


    async def aclose_processes(self, verbose: bool = True, timeout: float = 5.0):
        """
        Handles closing all processes
        """
        import anyio
        for func in self.on_close_funcs:
            with anyio.fail_after(timeout):
                try:
                    await func()
                    if verbose: self.logger.info(f"Called function {func.__name__} on close")
                except TimeoutError as e:
                    self.logger.error(f"Timeout calling function {func.__name__} on close: {e}")
                except Exception as e:
                    self.logger.trace(f"Error calling function {func.__name__} on close", error = e)
    
    def start_debug_mode(self):
        """
        Enters Debug Mode
        """
        self.logger.warning('Entering Debug Mode. Sleeping Forever')
        import time
        import sys
        while True:
            try:
                time.sleep(900)
            except Exception as e:
                self.logger.error(e)
                break
        sys.exit(0)


    async def arun_until_complete(
        self,
        termination_file: str = None,
    ):
        """
        Runs the event loop until complete
        """

        if termination_file is None:
            termination_file = os.getenv('WORKER_TERMINATION_FILE')
        
        from .types import GracefulKiller
        watch = GracefulKiller()
        tmp_kill_file = pathlib.Path(termination_file) if termination_file is not None else None
        while not watch.kill_now:
            try: 
                await asyncio.sleep(1.0)
                if tmp_kill_file is not None and tmp_kill_file.exists():
                    self.logger.warning(f"Found termination file: {tmp_kill_file}")
                    break
            except KeyboardInterrupt:
                self.logger.warning("Keyboard Interrupt")
                break
            except Exception as e:
                self.logger.error(f"Error: {e}")
                break
        
        self.end_all_processes()
        await self.aclose_processes()



    



# GlobalContext: GlobalContextClass = ProxyObject(GlobalContextClass)



from __future__ import annotations

"""
Kubernetes Context based on kr8s
"""
import os
import abc
import kr8s
import httpx
import base64
import functools
import contextlib
from typing import Optional, Dict, Any, List, Union, Generator, Literal, Type, AsyncGenerator, overload, TYPE_CHECKING
from .types import objects, aobjects
if TYPE_CHECKING:
    from kr8s import Api as ClientAPI
    from kr8s.asyncio import Api as AsyncClientAPI
    from .types import ObjectT, ObjectListT, aObjectT, aObjectListT



class KubernetesContext(abc.ABC):
    """
    A Single Kubernetes Context
    """

    def __init__(
        self, 
        name: Optional[str] = None, 
        url: Optional[str] = None,
        kubeconfig: Optional[str] = None,
        service_account: Optional[str] = None,
        context: Optional[str] = None,
        namespace: Optional[str] = None,
        **kwargs
    ):
        self.name = name
        self.url = url
        self.kubeconfig = kubeconfig
        self.service_account = service_account
        self.context = context
        self.namespace = namespace
        self._client: Optional[ClientAPI] = None
        self._aclient: Optional[AsyncClientAPI] = None
        self._ainitialized: bool = False
        self._kwargs = kwargs
        self.objects: Dict[str, Union['objects.APIObject', 'aobjects.APIObject']] = {}

    @property
    def client(self) -> ClientAPI:
        """
        Returns the client
        """
        if self._client is None:
            import kr8s
            self._client = kr8s.api(
                url = self.url,
                kubeconfig = self.kubeconfig,
                serviceaccount = self.service_account,
                context = self.context,
                namespace = self.namespace,
            )
        return self._client
    
    @property
    def aclient(self) -> AsyncClientAPI:
        """
        Returns the async client
        """
        if self._aclient is None:
            import kr8s
            self._aclient = kr8s.asyncio.api(
                url = self.url,
                kubeconfig = self.kubeconfig,
                serviceaccount = self.service_account,
                context = self.context,
                namespace = self.namespace,
            )
        return self._aclient
    

    def raise_if_not_ainit(self):
        """
        Raises if the async client is not initialized
        """
        if not self._ainitialized:
            raise ValueError("The async client is not initialized")
    
    async def ainit(self):
        """
        Initialize the async client
        """
        if not self._ainitialized:
            self._aclient = await self.aclient
            self._ainitialized = True


    @contextlib.asynccontextmanager
    async def call_api(
        self, 
        method: str = "GET",
        version: str = "v1",
        base: str = "",
        namespace: str = None,
        url: str = "",
        raise_for_status: bool = True,
        stream: bool = False,
        **kwargs,
    ) -> AsyncGenerator[httpx.Response, None]:
        """
        [Async] Call the API
        """
        await self.ainit()
        async with self.aclient.call_api(
            method = method,
            version = version,
            base = base,
            namespace = namespace,
            url = url,
            raise_for_status = raise_for_status,
            stream = stream,
            **kwargs
        ) as response:
            yield response
    
    @property
    def Pod(self) -> Type['objects.Pod']:
        """
        Returns the Pod object
        """
        if 'pod' not in self.objects:
            _pod = objects.Pod
            _pod.__init__ = functools.partialmethod(_pod.__init__, api = self.client)
            self.objects['pod'] = _pod
        return self.objects['pod']
    
    @property
    def aPod(self) -> Type['aobjects.Pod']:
        """
        Returns the async Pod object
        """
        self.raise_if_not_ainit()
        if 'apod' not in self.objects:
            _pod = aobjects.Pod
            _pod.__init__ = functools.partialmethod(_pod.__init__, api = self.aclient)
            self.objects['apod'] = _pod
        return self.objects['apod']
    

    @property
    def Service(self) -> Type['objects.Service']:
        """
        Returns the Service object
        """
        if 'service' not in self.objects:
            _service = objects.Service
            _service.__init__ = functools.partialmethod(_service.__init__, api = self.client)
            self.objects['service'] = _service
        return self.objects['service']
    
    @property
    def aService(self) -> Type['aobjects.Service']:
        """
        Returns the async Service object
        """
        self.raise_if_not_ainit()
        if 'aservice' not in self.objects:
            _service = aobjects.Service
            _service.__init__ = functools.partialmethod(_service.__init__, api = self.aclient)
            self.objects['aservice'] = _service
        return self.objects['aservice']
    

    @overload
    def get(
        self,
        name: Literal['pods'] = 'pods',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['objects.Pod']:
        ...

    @overload
    def get(
        self,
        name: Literal['services'] = 'services',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['objects.Service']:
        ...


    def get(
        self,
        name: str,
        *args,
        namespace: Optional[str] = kr8s.ALL,
        **kwargs,
    ) -> Union['ObjectT', 'ObjectListT']:
        """
        Get an object
        """
        return kr8s.get(name, *args, namespace = namespace, api = self.client, **kwargs)


    @overload
    async def aget(
        self,
        name: Literal['pods'] = 'pods',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['aobjects.Pod']:
        ...

    @overload
    async def aget(
        self,
        name: Literal['services'] = 'services',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['aobjects.Service']:
        ...

    @overload
    async def aget(
        self,
        name: Literal['deployments'] = 'deployments',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['aobjects.Deployment']:
        ...

    @overload
    async def aget(
        self,
        name: Literal['statefulsets'] = 'statefulsets',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['aobjects.StatefulSet']:
        ...

    @overload
    async def aget(
        self,
        name: Literal['configmaps'] = 'configmaps',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['aobjects.ConfigMap']:
        ...

    @overload
    async def aget(
        self,
        name: Literal['secrets'] = 'secrets',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['aobjects.Secret']:
        ...

    @overload
    async def aget(
        self,
        name: Literal['persistentvolumes'] = 'persistentvolumes',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['aobjects.PersistentVolume']:
        ...

    @overload
    async def aget(
        self,
        name: Literal['persistentvolumeclaims'] = 'persistentvolumeclaims',
        namespace: Optional[str] = kr8s.ALL,
        label_selector: Union[str, Dict] = None,
        field_selector: Union[str, Dict] = None,
        as_object: object = None,
        **kwargs,
    ) -> List['aobjects.PersistentVolumeClaim']:
        ...
    
    
    async def aget(
        self,
        name: str,
        *args,
        namespace: Optional[str] = kr8s.ALL,
        **kwargs,
    ) -> Union['aObjectT', 'aObjectListT']:
        """
        Get an object
        """
        await self.ainit()
        return await kr8s.asyncio.get(name, *args, namespace = namespace, api = self.aclient, **kwargs)



from __future__ import annotations

"""
Abstract Base Classes for Global Clients

- Requires `kvdb`
"""


import abc
import contextlib
from kvdb.io import cachify as _cachify
from typing import Optional, Type, TypeVar, Union, Set, List, Any, Dict, Literal, TYPE_CHECKING


if TYPE_CHECKING:
    from kvdb import KVDBSession, PersistentDict
    from lazyops.libs.logging import Logger
    from lazyops.libs.abcs.configs.base import AppSettings
    from lazyops.libs.abcs.types.api import BaseSchema
    from lazyops.libs.pooler import ThreadPool
    from lazyops.libs.abcs.clients import ClientTypes
    from lazyops.libs.abcs.configs.ctx import ComponentSchemaT

    ResultSchema = TypeVar('ResultSchema', bound = BaseSchema)

cachify = _cachify.create_context()

class BaseGlobalClient(abc.ABC):

    name: Optional[str] = None
    kind: Optional[str] = 'client'
    
    verbosity: Optional[int] = 0
    enable_registration: Optional[bool] = True

    extra_validators_enabled: Optional[bool] = False
    result_schema: Type['BaseSchema'] = None

    cache_enabled: Optional[bool] = True
    cachify_ttl: Optional[int] = None
    cachify_shared_global: Optional[bool] = False
    cachify_enabled: Optional[bool] = True
    
    serialization: Optional[str] = 'json'
    serialization_compression: Optional[str] = 'zstd'
    serialization_compression_level: Optional[int] = 19 # 6 for default (lz4, zlib) 19 for zstd
    data_expiration: Optional[int] = None
    data_shared_global: Optional[bool] = False
    
    cache_expiration: Optional[int] = 60 * 60 * 24 * 3 # 3 days
    cache_shared_global: Optional[bool] = False


    _logger: Optional['Logger'] = None
    _null_logger: Optional['Logger'] = None

    _settings: Optional['AppSettings'] = None
    _pooler: Optional['ThreadPool'] = None
    _kdb: Optional['KVDBSession'] = None

    _data: Optional['PersistentDict'] = None
    _cache: Optional['PersistentDict'] = None

    def __init__(self, *args, **kwargs):
        """
        Initializes the GlobalClient Object
        """
        self._kwargs = kwargs
        self.pre_init(**kwargs)
        self.cls_init(**kwargs)
        self.post_init(**kwargs)
        self.finalize_init(**kwargs)
        if self.enable_registration:
            self._register_client(*args, **kwargs)
    
    def pre_init(self, **kwargs):
        """
        Pre-Initializes the GlobalClient
        """
        pass

    def post_init(self, **kwargs):
        """
        Post-Initializes the GlobalClient
        """
        pass

    def cls_init(self, **kwargs):
        """
        Initializes the GlobalClient Class
        """
        pass
    
    def finalize_init(self, **kwargs):
        """
        Finalizes the GlobalClient Initialization
        """
        pass

    @property
    def logger(self) -> 'Logger':
        """
        Gets the logger
        """
        if self._logger is None:
            self._logger = self.settings.logger
        return self._logger
    
    @property
    def null_logger(self) -> 'Logger':
        """
        Gets the null logger
        """
        if self._null_logger is None:
            self._null_logger = self.settings.null_logger
        return self._null_logger
    

    @property
    def autologger(self) -> 'Logger':
        """
        Automatic Logger that is enabled in devel mode
        """
        return self.logger if (self.settings.is_development_env or self.verbosity >= 2) else self.null_logger

    
    @property
    def settings(self) -> 'AppSettings':
        """
        Gets the settings
        """
        if self._settings is None:
            from lazyops.libs.abcs.configs.lazy import get_module_settings
            self._settings = get_module_settings(self.__module__.__name__)
        return self._settings
    
    @property
    def pooler(self) -> 'ThreadPool':
        """
        Gets the pooler
        """
        if self._pooler is None:
            from lazyops.libs.pooler import ThreadPooler
            self._pooler = ThreadPooler
        return self._pooler
    
    @property
    def kdb(self) -> 'KVDBSession':
        """
        Gets the keydb
        """
        if self._kdb is None:
            self._kdb = self.settings.ctx.get_kdb_session(self.name)
        return self._kdb
    
    def pdict_get_extra_serialization_kwargs(self, name: str, **kwargs) -> Dict[str, Any]:
        """
        Gets the extra serialization kwargs
        """
        return {}
    
    @property
    def data(self) -> 'PersistentDict':
        """
        Gets the Persistent Dict for data
        """
        if self._data is None:
            serializer_kwargs = {
                'compression': self.serialization_compression,
                'compression_level': self.serialization_compression_level,
                'raise_errors': True,
                **self.pdict_get_extra_serialization_kwargs('data'),
            }
            base_key = f'{self.settings.ctx.module_name}.{self.name}.data' if \
                self.data_shared_global else \
                f'{self.settings.ctx.module_name}.{self.settings.app_env.name}.{self.name}.data'
        
            self._data = self.settings.ctx.get_persistent_dict(
                base_key = base_key,
                expiration = self.data_expiration,
                serializer = self.serialization,
                serializer_kwargs = serializer_kwargs,
            )
        return self._data


    @property
    def cache(self) -> 'PersistentDict':
        """
        Gets the Persistent Dict for cache
        """
        if self._cache is None:
            serializer_kwargs = {
                'compression': self.serialization_compression,
                'compression_level': self.serialization_compression_level,
                'raise_errors': True,
                **self.pdict_get_extra_serialization_kwargs('cache'),
            }
            base_key = f'{self.settings.ctx.module_name}.{self.name}.cache' if \
                self.cache_shared_global else \
                f'{self.settings.ctx.module_name}.{self.settings.app_env.name}.{self.name}.cache'
            self._cache = self.settings.ctx.get_persistent_dict(
                base_key = base_key,
                expiration = self.cache_expiration,
                serializer = self.serialization,
                serializer_kwargs = serializer_kwargs,
            )
        return self._cache


    

    @contextlib.contextmanager
    def catch(self, exc: Type['Exception'], msg: Optional[str] = None, fatal: Optional[bool] = False):
        """
        Catches an exception
        """
        try:
            yield
        except Exception as e:
            log = self.logger.trace if fatal else self.autologger.trace
            log(f'Handling Error: {msg}', e, prefix = self.name)
            if fatal: raise exc(msg, error = e) from e

    def run(
        self,
        data: 'BaseSchema',
        cachable: Optional[bool] = True,
        **kwargs,
    ) -> 'ResultSchema':
        """
        Runs the client
        """
        raise NotImplementedError("run must be implemented by the Client")
    
    async def arun(
        self,
        data: 'BaseSchema',
        cachable: Optional[bool] = True,
        **kwargs,
    ) -> 'ResultSchema':
        """
        Runs the client
        """
        raise NotImplementedError("arun must be implemented by the Client")
    
        
    def __call__(
        self,
        data: 'BaseSchema',
        cachable: Optional[bool] = True,
        is_async: Optional[bool] = True,
        **kwargs,
    ) -> 'ResultSchema':
        """
        Runs the client
        """
        method = self.arun if is_async else self.run
        return method(data, cachable = cachable, **kwargs)
    
    def cachify_validator_is_not_cachable(self, *args, cachable: Optional[bool] = True, **kwargs) -> bool:
        """
        Checks if the function is not cachable
        """
        from kvdb.io.cachify.helpers import is_not_cachable
        return is_not_cachable(*args, cachable = cachable, **kwargs)
    
    def cachify_validator_is_disabled(self, *args, disabled: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the function is disabled
        """
        from kvdb.io.cachify.helpers import is_disabled
        return is_disabled(*args, disabled = disabled, **kwargs)
    
    def cachify_validator_is_overwrite(self, *args, overwrite: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the function is overwrite
        """
        from kvdb.io.cachify.helpers import is_overwrite
        return is_overwrite(*args, overwrite = overwrite, **kwargs)
    
    def cachify_get_exclude_keys(self, func: str, **kwargs) -> List[str]:
        """
        Gets the exclude keys
        """
        return None
    
    def cachify_get_name_builder_kwargs(self, func: str, **kwargs) -> Dict[str, Any]:
        """
        Gets the name builder kwargs
        """
        return {}
    
    def cachify_get_extra_serialization_kwargs(self, func: str, **kwargs) -> Dict[str, Any]:
        """
        Gets the extra serialization kwargs
        """
        return {}

    def cachify_create_base_name(self, func: str, **kwargs) -> str:
        """
        Creates the base name
        """
        return f'{self.settings.ctx.module_name}.{self.name}' if self.cachify_shared_global  else \
            f'{self.settings.ctx.module_name}.{self.settings.app_env.name}.{self.name}'

    def validate_cachify(self, func: str, **kwargs) -> Dict[str, Any]:
        """
        Validates the cachify function
        """
        if not self.cachify_enabled: return None
        from .utils import create_cachify_build_name_func
        base_name = self.cachify_create_base_name(func, **kwargs)
        if 'name' not in kwargs: kwargs['name'] = create_cachify_build_name_func(
            base_name = base_name,
            **self.cachify_get_name_builder_kwargs(func, **kwargs),
        )
        if 'ttl' not in kwargs: kwargs['ttl'] = self.cachify_ttl
        if 'serializer' not in kwargs and 'encoder' not in kwargs and 'decoder' not in kwargs:
            kwargs['serializer'] = self.serialization
            kwargs['serializer_kwargs'] = {
                'compression': self.serialization_compression,
                'compression_level': self.serialization_compression_level,
                'raise_errors': True,
                **self.cachify_get_extra_serialization_kwargs(func, **kwargs),
            }
        if 'verbosity' not in kwargs and self.settings.is_local_env: kwargs['verbosity'] = 2
        kwargs['disabled'] = self.cachify_validator_is_not_cachable
        kwargs['overwrite_if'] = self.cachify_validator_is_overwrite
        kwargs['disabled_if'] = self.cachify_validator_is_disabled
        if exclude_keys := self.cachify_get_exclude_keys(func, **kwargs):
            kwargs['exclude_keys'] = exclude_keys
        kwargs['exclude_null'] = True
        return kwargs
    
    """
    Helper Methods
    """
    
    def _import(self, name: str) -> Any:
        """
        Imports a module
        """
        from lazyops.utils.lazy import lazy_import
        return lazy_import(name)
    

    def _register_client(self, kind: Optional[str] = None, include_kind: Optional[bool] = None, **kwargs):
        """
        Registers the client
        """
        kind = kind or self.kind
        self.settings.ctx.register_client(self, kind = self.kind, include_kind = include_kind, **kwargs)

    def _get_client(
        self, 
        name: str,
        state: Optional[Literal['global', 'local']] = None,
        kind: Optional[str] = 'client',
        include_kind: Optional[bool] = None,
        **kwargs,
    ) -> 'ClientTypes':
        """
        Retrieves a client
        """
        return self.settings.ctx.get_client(name, state = state, kind = kind, include_kind = include_kind, **kwargs)

    def _get_component_client(
        self,
        name: str,
        *parts: str,
        kind: Optional[str] = None,
        include_kind: Optional[bool] = None,
    ) -> 'ClientTypes':
        """
        Gets a component client
        """
        return self.settings.ctx.get_component_client(name, *parts, kind = kind, include_kind = include_kind)
    

    def _get_component_schema(
        self,
        name: str,
        *parts: str,
        kind: Optional[str] = None,
        include_kind: Optional[bool] = None,
    ) -> Type['ComponentSchemaT']:
        """
        Gets a component schema
        """
        return self.settings.ctx.get_component_schema(name, *parts, kind = kind, include_kind = include_kind)
    
    def _register_component_client(
        self,
        *parts: str,
        kind: Optional[str] = None,
        include_kind: Optional[bool] = None,
    ):
        """
        Registers a component client
        """
        return self.settings.ctx.register_component_client(self, *parts, kind = kind, include_kind = include_kind)
    


from __future__ import annotations

import os
import abc
from lazyops.libs.pooler import ThreadPooler
from typing import Optional, List, Dict, Any, Union, Type, Tuple, Generator, TYPE_CHECKING
from lazyops.imports._gspread import resolve_gspread
resolve_gspread(required = True)

import gspread


class WorksheetContext(abc.ABC):
    """
    The Current Worksheet Context
    """
    def __init__(
        self,
        gc: 'gspread.Client',
        url: str,
    ):
        """
        Initializes the Worksheet Context
        """
        self.gc = gc
        self.url = url
        self.sheet = self.gc.open_by_url(self.url)
        self.name = self.sheet.title
        self.sheet_name: self.sheet_name = self.sheet.sheet1.title
        self.wks = self.sheet.worksheet(self.sheet_name)
        self.header_index: Dict[str, Dict[str, str]] = {
            self.sheet_name: self.wks.row_values(1)
        }
        # self.headers = self.wks.row_values(1)
        self._worksheet_names: Optional[List[str]] = None

    @property
    def worksheet_names(self) -> List[str]:
        """
        Return the list of worksheet names
        """
        if self._worksheet_names is None:
            self._worksheet_names = [sheet.title for sheet in self.sheet.worksheets(False)]
        return self._worksheet_names
    
    @property
    def headers(self):
        """
        Returns the current sheet headers
        """
        return self.header_index[self.sheet_name]
    
    def duplicate_worksheet(self, name: str, src_name: Optional[str] = None) -> 'gspread.worksheet.Worksheet':
        """
        Duplicate the current worksheet
        """
        src_name = src_name or self.sheet_name
        wks = self.sheet.worksheet(src_name)
        self._worksheet_names = None
        return wks.duplicate(new_sheet_name = name)
        # self.header_index[name] = new_wks.row_values(1)
        # return new_wks
    
    async def aduplicate_worksheet(self, name: str, src_name: Optional[str] = None) -> 'gspread.worksheet.Worksheet':
        """
        Duplicate a worksheet
        -> _template -> name
        """
        src_name = src_name or self.sheet_name
        wks = self.sheet.worksheet(src_name)
        self._worksheet_names = None
        return await ThreadPooler.run_async(wks.duplicate, new_sheet_name = name)
        # self.header_index[name] = new_wks.row_values(1)
        # return new_wks

    def get_worksheet(self, key: Optional[Union[str, int]] = None) -> 'gspread.worksheet.Worksheet':
        """
        Returns a sheet by name or index
        """
        if key is None: key = self.sheet_name
        if key == self.sheet_name: return self.wks
        if isinstance(key, int):
            return self.sheet.worksheet(self.worksheet_names[key])
            # sn = self.worksheet_names[key]
            # wks = self.sheet.worksheet(self.worksheet_names[key])
            # if sn not in self.header_index:
            #     self.header_index[sn] = wks.row_values(1)
            # return wks
        if isinstance(key, str) and key not in self.worksheet_names:
            return self.duplicate_worksheet(key)
        return self.sheet.worksheet(key)
        # if key not in self.header_index:
        #     self.header_index[key] = wks.row_values(1)
        # return wks

    async def aget_worksheet(self, key: Union[str, int]) -> 'gspread.worksheet.Worksheet':
        """
        Returns a sheet by name or index
        """
        if key is None: key = self.sheet_name
        if key == self.sheet_name: return self.wks
        if isinstance(key, int):
            return await ThreadPooler.run_async(self.sheet.worksheet, self.worksheet_names[key])
            # sn = self.worksheet_names[key]
            # wks = await ThreadPooler.run_async(self.sheet.worksheet, self.worksheet_names[key])
            # if sn not in self.header_index:
            #     self.header_index[sn] = wks.row_values(1)
            # return wks
        if isinstance(key, str) and key not in self.worksheet_names:
            return await self.aduplicate_worksheet(key)
        return await ThreadPooler.run_async(self.sheet.worksheet, key)
        # wks = await ThreadPooler.run_async(self.sheet.worksheet, key)
        # if key not in self.header_index:
        #     self.header_index[key] = wks.row_values(1)
        # return wks

    def append_row(self, data: List[Any], sheet_name: Optional[str] = None) -> None:
        """
        Append a row into the worksheet
        """
        if isinstance(data, dict): data = list(data.values())
        wks = self.get_worksheet(sheet_name)
        wks.append_row(data)

    async def aappend_row(
        self, 
        data: Union[Dict[str, Any], List[Any]],
        sheet_name: Optional[str] = None,
        **kwargs
    ):
        """
        Append a row into the worksheet
        """
        if isinstance(data, dict): data = list(data.values())
        # wks = self[sheet_name]
        wks = await self.aget_worksheet(sheet_name)
        await ThreadPooler.run_async(wks.append_row, data)

    def append_rows(self, data: List[List[Any]], sheet_name: Optional[str] = None) -> None:
        """
        Append rows into the worksheet
        """
        if not isinstance(data, list) or not isinstance(data[0], (list, dict)):
            return self.append_row(data, sheet_name = sheet_name)
        data = [list(d.values()) if isinstance(d, dict) else d for d in data]
        wks = self.get_worksheet(sheet_name)
        wks.append_rows(data)
    
    async def aappend_rows(
        self, 
        data: Union[List[Dict[str, Any]], List[List[Any]]],
        sheet_name: Optional[str] = None,
        **kwargs
    ):
        """
        Append rows into the worksheet
        """
        if not isinstance(data, list) or not isinstance(data[0], (list, dict)):
            return await self.aappend_row(data, sheet_name = sheet_name)
        data = [list(d.values()) if isinstance(d, dict) else d for d in data]
        wks = await self.aget_worksheet(sheet_name)
        await ThreadPooler.run_async(wks.append_rows, data)

    
    def insert_row(
        self,
        data: Union[Dict[str, Any], List[Any]],
        index: int= 1,
        value_input_option: str = 'raw',
        inherit_from_before: bool = True,
        sheet_name: Optional[str] = None,
    ):
        """
        Insert a row into the worksheet
        """
        if isinstance(data, dict): data = list(data.values())
        wks = self.get_worksheet(sheet_name)
        wks.insert_row(data, index = index, value_input_option = value_input_option, inherit_from_before = inherit_from_before)

    async def ainsert_row(
        self,
        data: Union[Dict[str, Any], List[Any]],
        index: int= 1,
        value_input_option: str = 'raw',
        inherit_from_before: bool = True,
        sheet_name: Optional[str] = None,
    ):
        """
        Insert a row into the worksheet
        """
        if isinstance(data, dict): data = list(data.values())
        wks = await self.aget_worksheet(sheet_name)
        await ThreadPooler.run_async(wks.insert_row, data, index = index, value_input_option = value_input_option, inherit_from_before = inherit_from_before)

    def insert_rows(
        self,
        data: Union[List[Dict[str, Any]], List[List[Any]]],
        index: int= 1,
        value_input_option: str = 'raw',
        inherit_from_before: bool = True,
        sheet_name: Optional[str] = None,
    ):
        """
        Insert rows into the worksheet
        """
        if not isinstance(data, list) or not isinstance(data[0], (list, dict)):
            return self.insert_row(data, index = index, value_input_option = value_input_option, inherit_from_before = inherit_from_before, sheet_name = sheet_name)
        data = [list(d.values()) if isinstance(d, dict) else d for d in data]
        wks = self.get_worksheet(sheet_name)
        wks.insert_rows(data, index = index, value_input_option = value_input_option, inherit_from_before = inherit_from_before)
    
    async def ainsert_rows(
        self,
        data: Union[List[Dict[str, Any]], List[List[Any]]],
        index: int= 1,
        value_input_option: str = 'raw',
        inherit_from_before: bool = True,
        sheet_name: Optional[str] = None,
    ):
        """
        Insert rows into the worksheet
        """
        if not isinstance(data, list) or not isinstance(data[0], (list, dict)):
            return await self.ainsert_row(data, index = index, value_input_option = value_input_option, inherit_from_before = inherit_from_before, sheet_name = sheet_name)
        data = [list(d.values()) if isinstance(d, dict) else d for d in data]
        wks = await self.aget_worksheet(sheet_name)
        await ThreadPooler.run_async(wks.insert_rows, data, index = index, value_input_option = value_input_option, inherit_from_before = inherit_from_before)

    def get_row(self, index: int, sheet_name: Optional[str] = None, raw_values: Optional[bool] = False) -> Union[List[Any], Dict[str, Any]]:
        """
        Get a row from the worksheet
        """
        wks = self.get_worksheet(sheet_name)
        values = wks.row_values(index)
        if raw_values: return values
        if wks.title not in self.header_index:
            self.header_index[wks.title] = wks.row_values(1)
        return dict(zip(self.header_index[wks.title], values))
    
    async def aget_row(self, index: int, sheet_name: Optional[str] = None, raw_values: Optional[bool] = False) -> Union[List[Any], Dict[str, Any]]:
        """
        Get a row from the worksheet
        """
        wks = await self.aget_worksheet(sheet_name)
        values = await ThreadPooler.run_async(wks.row_values, index)
        if raw_values: return values
        if wks.title not in self.header_index:
            self.header_index[wks.title] = wks.row_values(1)
        return dict(zip(self.header_index[wks.title], values))
    
    def update_row(self, data: Union[Dict[str, Any], List[Any]], index: Optional[int] = None, sheet_name: Optional[str] = None) -> None:
        """
        Update a row in the worksheet
        """
        if isinstance(data, dict): 
            if index is None and 'idx' in data: index = data.pop('idx')
            data = list(data.values())
        assert index is not None, 'Index must be provided'
        wks = self.get_worksheet(sheet_name)
        wks.update(range_name=f'A{index}', values = [data])

    async def aupdate_row(
        self, 
        data: Union[Dict[str, Any], List[Any]],
        index: Optional[int] = None,
        sheet_name: Optional[str] = None,
        **kwargs
    ):
        """
        Update a row in the worksheet
        """
        if isinstance(data, dict): 
            if index is None and 'idx' in data: index = data.pop('idx')
            data = list(data.values())
        assert index is not None, 'Index must be provided'
        wks = await self.aget_worksheet(sheet_name)
        await ThreadPooler.run_async(wks.update, range_name=f'{index}:{index}', values = data)

    def resize_height(self, height: int, wks: Optional[gspread.worksheet.Worksheet] = None) -> None:
        """
        Resize the height of the worksheet
        """
        if not wks: wks = self.wks
        sheet_id = wks._properties['sheetId']
        body = {
            "requests": [
                {
                    "updateDimensionProperties": {
                        "range": {
                            "sheetId": sheet_id, "dimension": "ROWS",
                            "startIndex": 0, "endIndex": wks.row_count
                        },
                        "properties": {"pixelSize": height}, "fields": "pixelSize"
                    }
                }
            ]
        }
        wks.batch_update(body)
    
    async def aresize_height(
        self, 
        height: int = 22,
        wks: Optional['gspread.worksheet.Worksheet'] = None,
    ):
        """
        Resize the height of the worksheet
        """
        # https://stackoverflow.com/questions/57177998/gspread-column-sizing
        if not wks: wks = self.wks
        sheet_id = wks._properties['sheetId']
        body = {
            "requests": [
                {
                    "updateDimensionProperties": {
                        "range": {
                            "sheetId": sheet_id, "dimension": "ROWS",
                            "startIndex": 0, "endIndex": wks.row_count
                        },
                        "properties": {"pixelSize": height}, "fields": "pixelSize"
                    }
                }
            ]
        }
        return await ThreadPooler.run_async(wks.batch_update, body)

    def get_items(self, sheet_name: Optional[str] = None, include_idx: Optional[bool] = True) -> Dict[int, Any]:
        """
        Gets all the items in the worksheet
        """
        wks = self.get_worksheet(sheet_name)
        headers = wks.row_values(1)
        results = {}
        for n, row in enumerate(wks.get()):
            if n == 0: continue
            if len(row) != len(headers):  row.extend([None] * (len(headers) - len(row)))
            data = dict(zip(headers, row))
            if include_idx: data['idx'] = n + 1
            results[n + 1] = data
        return results
    
    def items(self, sheet_name: Optional[str] = None, include_idx: Optional[bool] = True) -> Generator[Tuple[int, Dict[str, Any]], None, None]:
        """
        Returns a tuple of (idx, item)
        """
        wks = self.get_worksheet(sheet_name)
        if wks.title not in self.header_index:
            self.header_index[wks.title] = wks.row_values(1)
        headers = self.header_index[wks.title]
        for n, row in enumerate(wks.get(major_dimension='ROWS')):
            if n == 0: continue
            if len(row) != len(headers):  row.extend([None] * (len(headers) - len(row)))
            data = dict(zip(headers, row))
            if include_idx: data['idx'] = n + 1
            yield n + 1, data
        
    def __setitem__(self, index: int, value: Union[Dict[str, Any], List[Any]]):
        """
        Set an item in the worksheet
        """
        self.update_row(value, index = index)

    def __getitem__(self, index: int) -> Union[Dict[str, Any], List[Any]]:
        """
        Get an item from the worksheet
        """
        return self.get_row(index)
    
    def __len__(self) -> int:
        """
        Get the length of the worksheet
        """
        return self.wks.row_count


class WorksheetManager(abc.ABC):
    """
    The Worksheet Manager
    """
    def __init__(
        self,
        adc: Optional[str] = None,
        url: Optional[str] = None,
        **kwargs,
    ):
        """
        The Worksheet Manager
        """
        self.adc = adc or os.getenv('GSPREAD_ADC', os.getenv('GOOGLE_APPLICATION_CREDENTIALS', None))
        self._gc: Optional[gspread.Client] = None
        self.default_url: Optional[str] = url
        self.url_to_ctx: Dict[str, str] = {} 
        self.ctxs: Dict[str, WorksheetContext] = {}
        self._ctx: Optional[WorksheetContext] = None
    
    @property
    def gc(self) -> gspread.Client:
        """
        Returns the gspread client
        """
        if self._gc is None:
            if self.adc: 
                self._gc = gspread.service_account(filename = self.adc)
            else: self._gc = gspread.oauth()
            self._gc.http_client.set_timeout(timeout = 60 * 60)
        return self._gc
    

    @property
    def ctx(self) -> WorksheetContext:
        """
        The default context
        """
        if self._ctx is None:
            self.init_ctx()
        return self._ctx

    def init_ctx(
        self, 
        name: Optional[str] = None,
        url: Optional[str] = None,
        set_as_default: Optional[bool] = False,
    ):
        """
        Initializes the context
        """
        name = name or "default"
        if name not in self.ctxs:
            url = url or self.default_url
            assert url, 'No URL provided'
            self.ctxs[name] = WorksheetContext(self.gc, url)
            self.url_to_ctx[url] = name
        if self._ctx is None or set_as_default: self._ctx = self.ctxs[name]
    
    def get_ctx(
        self, 
        name: Optional[str] = None,
        url: Optional[str] = None,
        set_as_default: Optional[bool] = False,
        **kwargs
    ) -> WorksheetContext:
        """
        Returns a context
        """
        if name is None or name == "default": return self.ctx
        self.init_ctx(name = name, url = url, set_as_default = set_as_default, **kwargs)
        return self.ctxs[name]
    



from __future__ import annotations

"""
User Data Flows
"""
import time
from ..types.properties import StatefulProperty
from ..types.user_data import AZUserData
from .admin import AZManagementClient
from typing import List, Optional, Any, TYPE_CHECKING

class UserDataFlow(StatefulProperty[AZUserData]):
    """
    User Data Flow
    """
    name: Optional[str] = "user_data"
    schema_type = AZUserData

    cache_key_prefix: Optional[str] = None

    def __init__(
        self, 
        user_id: str,
        **kwargs,
    ):
        self.is_client_id: Optional[bool] = False
        if '@clients' in user_id: 
            user_id = user_id.split('@clients')[0]
            self.is_client_id = True
        cache_key = self.settings.create_cache_key(
            prefix = self.cache_key_prefix,
            suffix = user_id,
            include_client_id = True,
            # kind = 'user_data',
        )
        super().__init__(
            cache_key = cache_key,
            **kwargs,
        )
        self.user_id = user_id
        self.mtg = AZManagementClient

    def load(self) -> AZUserData:
        """
        Retrieves the User Data
        """
        if self.is_client_id:
            return AZUserData(user_id = self.user_id, user_metadata = {}, app_metadata = {})
        response = self.mtg.hget(f'users/{self.user_id}')
        result = self.mtg._process_response(response)
        return AZUserData(**result)

    async def aload(self) -> AZUserData:
        """
        Loads the User Data
        """
        if self.is_client_id:
            return AZUserData(user_id = self.user_id, user_metadata = {}, app_metadata = {})
        response = await self.mtg.ahget(f'users/{self.user_id}')
        result = self.mtg._process_response(response)
        return AZUserData(**result)
    
    def get_ttl(self, data: AZUserData) -> int:
        """
        Returns the TTL
        """
        return None if self.is_client_id else self.settings.user_data_expiration or 60
        # return data.expiration_ts - int(time.time())

    def is_expired(self, data: AZUserData) -> bool:
        """
        Returns True if the Data is Expired
        """
        # self.settings.autologger.info(f'Checking if User Data is Expired: {self.user_id}: {data.is_expired}')
        return False if self.is_client_id else data.is_expired
    
    @property
    def user_data(self) -> AZUserData:
        """
        Returns the User Data
        """
        return self.resource
    
    @property
    async def auser_data(self) -> AZUserData:
        """
        Returns the User Data
        """
        return await self.aresource

    if TYPE_CHECKING:
        
        @classmethod
        def get(cls, user_id: str) -> AZUserData:
            """
            Get the User Data
            """
            ...
        
        @classmethod
        async def aget(cls, user_id: str) -> AZUserData:
            """
            Get the User Data
            """
            ...

        def fetch(self) -> AZUserData:
            """
            Fetches the User Data
            """
            ...

        async def afetch(self) -> AZUserData:
            """
            Fetches the User Data
            """
            ...

        @property
        def resource(self) -> AZUserData:
            """
            Returns the Object
            """
            ...

        @property
        async def aresource(self) -> AZUserData:
            """
            Returns the Object Resource
            """
            ...




from typing import Union, TypeVar, Any, TYPE_CHECKING


if TYPE_CHECKING:
    from .base import BaseGlobalClient
    from .http import BaseAPIClient, HTTPPoolClient
    from .browser import BrowserClient

    from ..sql.database.base import DatabaseClientBase

    from async_openai.manager import OpenAIManager
    from kvdb.tasks.base import BaseTaskWorker

    GlobalClientT = TypeVar('GlobalClientT', bound = BaseGlobalClient)
    APIClientT = TypeVar('APIClientT', bound = BaseAPIClient)

    GlobalClientType = Union[
        GlobalClientT,
        APIClientT,
        BrowserClient,
        OpenAIManager,
        HTTPPoolClient,
        DatabaseClientBase,
        Any
    ]
    
    LocalClientType = Union[
        BaseTaskWorker,
        Any
    ]
    ClientTypes = Union[
        GlobalClientType,
        LocalClientType,
    ]



from __future__ import annotations

"""
Client Tokens and App Token Flows
"""

from lazyops.libs import lazyload
from lazyops.libs.abcs.configs.types import AppEnv
from ..types.properties import StatefulProperty
from ..types.tokens import AccessToken
from ..utils.lazy import get_az_settings, logger
from ..utils.helpers import normalize_audience_name
from typing import List, Optional, Any, Dict, Union, TYPE_CHECKING

if lazyload.TYPE_CHECKING:
    import niquests
else:
    niquests = lazyload.LazyLoad("niquests")

class BaseTokenFlow(StatefulProperty[AccessToken]):
    """
    Base Token Flow
    """
    name: Optional[str] = None
    schema_type = AccessToken
    
    def __init__(
        self, 
        audience: str,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        oauth_url: Optional[str] = None,
        **kwargs,
    ):
        """
        Initializes the Token Flow
        """
        cache_key = kwargs.pop('cache_key', None)
        cache_key = cache_key or self.settings.create_cache_key(audience)
        super().__init__(
            cache_key = cache_key,
            **kwargs,
        )
        self.audience = audience
        self.client_id = client_id or self.settings.client_id
        self.client_secret = client_secret or self.settings.client_secret
        self.oauth_url = oauth_url or self.settings.oauth_url

    
    @property
    def access_token(self) -> AccessToken:
        """
        Returns the Access Token
        """
        return self.resource
    
    @property
    def token(self) -> str:
        """
        Returns the Access Token
        """
        return self.resource.access_token
    
    @property
    async def atoken(self) -> str:
        """
        Returns the Access Token
        """
        return (await self.aresource).access_token
    
    @property
    def auth_headers(self) -> Dict[str, str]:
        """
        Returns the Auth Headers
        """
        return {'Authorization': f'Bearer {self.token}'}
    
    @property
    async def aauth_headers(self) -> Dict[str, str]:
        """
        Returns the Auth Headers
        """
        return {'Authorization': f'Bearer {await self.atoken}'}
    
    
    def load(self) -> AccessToken:
        """
        Retrieves the Access Token from the API
        """
        data = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'audience': self.audience,
            'grant_type': 'client_credentials',
        }
        # logger.info(f'Retrieving Access Token from {self.oauth_url}')
        response = niquests.post(self.oauth_url, data = data)
        response.raise_for_status()
        return AccessToken(**response.json())
    
    async def aload(self) -> AccessToken:
        """
        Retrieves the Access Token from the API
        """
        data = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'audience': self.audience,
            'grant_type': 'client_credentials',
        }
        # logger.info(f'Retrieving Access Token from {self.oauth_url}')
        async with niquests.AsyncSession() as s:
            response = await s.post(self.oauth_url, data = data)
            response.raise_for_status()
            return AccessToken(**response.json())


    def get_ttl(self, data: AccessToken) -> int:
        """
        Returns the TTL
        """
        return data.expires_in - 60
    

    def is_expired(self, data: AccessToken) -> bool:
        """
        Returns True if the Data is Expired
        """
        return data.is_expired


    if TYPE_CHECKING:
        def fetch(self) -> AccessToken:
            """
            Fetches the Access Token
            """
            ...

        async def afetch(self) -> AccessToken:
            """
            Fetches the Access Token
            """
            ...
        

        @classmethod
        def get(
            cls,
            audience: str,
            client_id: Optional[str] = None,
            client_secret: Optional[str] = None,
            oauth_url: Optional[str] = None,
            **kwargs,
        ) -> AccessToken:
            """
            Get the Access Token
            """
            ...
        

        @classmethod
        async def aget(
            cls,
            audience: str,
            client_id: Optional[str] = None,
            client_secret: Optional[str] = None,
            oauth_url: Optional[str] = None,
            **kwargs,
        ) -> AccessToken:
            """
            Get the Access Token
            """
            ...

        @property
        def resource(self) -> AccessToken:
            """
            Returns the Object
            """
            ...

        @property
        async def aresource(self) -> AccessToken:
            """
            Returns the Object Resource
            """
            ...



class ClientCredentialsFlow(BaseTokenFlow):
    """
    Can be used to get a token for any API using Client Credentials flow.
    Specify the API with the `audience` param.
    """
    name: Optional[str] = "client_token"
    schema_type = AccessToken



class APIClientCredentialsFlow(BaseTokenFlow):
    """
    Can be used to get a token for any API using Client Credentials flow.
    Specify the API with the `audience` param.
    """
    name: Optional[str] = "api_token"
    schema_type = AccessToken

    def __init__(
        self, 
        endpoint: str, 
        api_client_id: Optional[str] = None, 
        api_client_env: Optional[Union[str, AppEnv]] = None, 
        audience: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        oauth_url: Optional[str] = None,
        **kwargs,
    ):
        """
        Initializes the API Client Credentials for the Audience
        """
        # key = endpoint.split('://', 1)[-1].replace('/', '_')
        key = normalize_audience_name(endpoint)
        if api_client_id: key += f'.{api_client_id}'
        if api_client_env: 
            if isinstance(api_client_env, AppEnv): api_client_env = api_client_env.name
            key += f'.{api_client_env}'
        if audience: key += f'.{normalize_audience_name(audience)}'
        super().__init__(
            audience = audience or self.settings.management_api_url,
            client_id = client_id or self.settings.client_id,
            client_secret = client_secret or self.settings.client_secret,
            oauth_url = oauth_url or self.settings.oauth_url,
            cache_key = key,
            **kwargs,
        )
        self.endpoint = endpoint

    """
    Might need to rework these later
    """

    def load_data(self) -> Dict[str, Any]:
        """
        Loads the Client Data
        """
        return self.pdict.get(self.data_cache_key, {})

    async def aload_data(self) -> Dict[str, Any]:
        """
        Loads the Client Data
        """
        return await self.pdict.aget(self.data_cache_key, {})
    
    def save_data(self, data: Dict[str, Any], ex: Optional[int] = None):
        """
        Saves the Data
        """
        self.pdict.set(self.data_cache_key, data, ex = ex)

    async def asave_data(self, data: Dict[str, Any], ex: Optional[int] = None):
        """
        Saves the Data
        """
        await self.pdict.aset(self.data_cache_key, data, ex = ex)
    



from typing import TYPE_CHECKING, Union, Type

if TYPE_CHECKING:
    from .admin import AZManagementClient, AZManagementAPI
    from .api_keys import APIKeyDataFlow
    from .tokens import ClientCredentialsFlow, APIClientCredentialsFlow
    from .user_data import UserDataFlow
    from .user_session import UserSessionFlow

    AZFlow = Union[
        APIClientCredentialsFlow,
        APIKeyDataFlow,
        ClientCredentialsFlow,
        UserDataFlow,
        UserSessionFlow,
    ]

    AZFlowSchema = Type[
        Union[
            APIClientCredentialsFlow,
            APIKeyDataFlow,
            AZManagementAPI,
            ClientCredentialsFlow,
            UserDataFlow,
            UserSessionFlow,
        ]
    ]



from __future__ import annotations

"""
Wrapper for niquest Sessions
"""

from abc import ABC
from typing import Optional, List, Dict, Any, Union
from lazyops.libs import lazyload


if lazyload.TYPE_CHECKING:
    import niquests
    from niquests import Session, AsyncSession, Response, AsyncResponse
else:
    niquests = lazyload.LazyLoad("niquests")


class RequestSessionContext(ABC):
    """
    The Request Session Context

    - This is a wrapper for the niquests Session, which will maintain handling of long running sessions
    """

    def __init__(self):
        
        self.session: Optional['Session'] = None
        self.asession: Optional['AsyncSession'] = None

        self.n_requests: int = 0
        self.an_requests: int = 0
    






from __future__ import annotations

"""
FastAPI Security Types for Auth0
"""

from typing import Optional, Annotated
from fastapi import Request, Security
from fastapi.security.api_key import  APIKeyQuery, APIKeyBase, APIKeyIn, APIKeyHeader
from fastapi.security.api_key import APIKey as BaseAPIKey
from fastapi.security.http import HTTPBase, HTTPBasic, HTTPBasicCredentials, HTTPBearer, HTTPBearerModel, HTTPAuthorizationCredentials
from . import errors

APIKeyScheme = APIKeyHeader(
    name = "x-api-key",
    scheme_name = "API Key",
    description = "API Key Authentication",
    auto_error = False,
)
AuthorizationScheme = HTTPBase(
    scheme = "bearer",
    scheme_name = "Authorization",
    description = "Bearer Token Authentication",
    auto_error = False,
)

APIKey = Annotated[Optional[str], Security(APIKeyScheme)]
Authorization = Annotated[Optional[HTTPAuthorizationCredentials], Security(AuthorizationScheme)] 



from __future__ import annotations

"""
A Generic Postgres SQL Database Client

This module provides an abstract base class for a generic Postgres SQL database client.

Requires `sqlalchemy` and `psycopg2` to be installed.
"""
import gc
import abc
import asyncio
import datetime
import contextlib
from pathlib import Path
from pydantic.networks import PostgresDsn
from pydantic_settings import BaseSettings
from pydantic import validator, model_validator, computed_field, BaseModel, Field, PrivateAttr

from sqlalchemy import text as sql_text, TextClause
from sqlalchemy.pool import NullPool, AsyncAdaptedQueuePool
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.ext.asyncio import AsyncSession, AsyncEngine

from lazyops.utils.logs import logger, Logger
from lazyops.utils.lazy import lazy_import
from ...utils.helpers import update_dict
from typing import Any, Dict, List, Optional, Type, Literal, Iterable, Tuple, TypeVar, Union, Annotated, Callable, Generator, AsyncGenerator, Set, TYPE_CHECKING

if TYPE_CHECKING:
    from sqlalchemy.orm import DeclarativeBase
    from sqlalchemy.engine import Result
    from lazyops.libs.pooler import ThreadPool
    from lazyops.libs.abcs.configs.base import AppSettings
    from .templates import SQLTemplates
    from .types import ModelType


class PostgresConfig(BaseModel):

    """
    The Postgres Settings
    """
    url: Optional[PostgresDsn] = None
    readonly_url: Optional[PostgresDsn] = None
    superuser_url: Optional[PostgresDsn] = None

    engine_poolclass: Optional[Union[Type[Union[NullPool, AsyncAdaptedQueuePool]], str]] = 'sqlalchemy.pool.AsyncAdaptedQueuePool'
    engine_json_serializer: Optional[Union[Callable, str]] = 'json'

    engine_kwargs: Optional[Dict[str, Any]] = Field(default_factory = dict)
    engine_rw_kwargs: Optional[Dict[str, Any]] = Field(default_factory = dict)
    engine_ro_kwargs: Optional[Dict[str, Any]] = Field(default_factory = dict)

    session_kwargs: Optional[Dict[str, Any]] = Field(default_factory = dict)
    session_ro_kwargs: Optional[Dict[str, Any]] = Field(default_factory = dict)
    session_rw_kwargs: Optional[Dict[str, Any]] = Field(default_factory = dict)

    @computed_field
    @property
    def url_password(self) -> Optional[str]:
        """
        The Password
        """
        if ':' in str(self.url) and '@' in str(self.url):
            return str(self.url).split('@')[0].split(':')[-1]
        return None

    @computed_field
    @property
    def superuser_url_password(self) -> Optional[str]:
        """
        The Password
        """
        if not self.superuser_url: return None
        if ':' in str(self.superuser_url) and '@' in str(self.superuser_url):
            return str(self.superuser_url).split('@')[0].split(':')[-1]
        return None
    
    @computed_field
    @property
    def safe_url(self) -> Optional[str]:
        """
        The Password
        """
        if not self.url: return None
        return str(self.url).replace(self.url_password, '********') if self.url_password else str(self.url)
    
    @computed_field
    @property
    def safe_superuser_url(self) -> Optional[str]:
        """
        The Password
        """
        if not self.superuser_url: return None
        return str(self.superuser_url).replace(self.superuser_url_password, '********') if self.superuser_url_password else str(self.superuser_url)
    

    def log_db_url(self, verbose: Optional[bool] = False):
        """
        Log the Database URL
        """
        if not verbose: return
        logger.info(f'Using URL: |g|{self.safe_url}|e|', colored = True, prefix = 'PostgresDB')

    def log_readonly_warning(self, verbose: Optional[bool] = False):
        """
        Log the Readonly Warning
        """
        if not verbose: return
        if not self.readonly_url:
            safe_url = self.safe_url
            logger.info(f'|y|Readonly URL not set|e|, using default URL: {safe_url}', colored = True, prefix = 'PostgresDB')


    @model_validator(mode = 'after')
    def post_init_validate(self):
        """
        Validate after init
        """
        if not self.superuser_url: self.superuser_url = self.url
        if isinstance(self.engine_poolclass, str):
            self.engine_poolclass = lazy_import(self.engine_poolclass)
        if isinstance(self.engine_json_serializer, str):
            if self.engine_json_serializer == 'json':
                try:
                    from kvdb.io.serializers import get_serializer
                    serializer = get_serializer('json')
                    self.engine_json_serializer = serializer.dumps
                except ImportError:
                    from lazyops.utils.serialization import Json
                    self.engine_json_serializer = Json.dumps
            else:
                try:
                    self.engine_json_serializer = lazy_import(self.engine_json_serializer)
                except Exception as e:
                    logger.error(f'Failed to import the JSON Serializer: {e}')
        if not self.engine_kwargs:
            self.engine_kwargs['pool_pre_ping'] = True
        if not self.session_rw_kwargs:
            self.session_rw_kwargs['expire_on_commit'] = False
        if not self.session_ro_kwargs:
            self.session_ro_kwargs['autoflush'] = False
            self.session_ro_kwargs['autocommit'] = False
        return self
    
    @classmethod
    def from_settings(cls, settings: 'BaseSettings') -> 'PostgresConfig':
        """
        Create from Settings
        """
        return cls.model_validate(settings, from_attributes = True)
    

    def get_engine_kwargs(self, readonly: Optional[bool] = False, verbose: Optional[bool] = False, **engine_kwargs) -> Dict[str, Any]:
        """
        Get the Engine KWargs
        """
        kwargs = self.engine_kwargs or {}
        if engine_kwargs: kwargs = update_dict(kwargs, engine_kwargs)
        kwargs = (
            update_dict(kwargs, self.engine_ro_kwargs)
            if readonly
            else update_dict(kwargs, self.engine_rw_kwargs)
        )
        self.log_readonly_warning(verbose) if readonly else self.log_db_url(verbose)
        kwargs['url'] = str((self.readonly_url or self.url) if readonly else self.url)
        if self.engine_json_serializer:
            kwargs['json_serializer'] = self.engine_json_serializer
        if self.engine_poolclass:
            kwargs['poolclass'] = self.engine_poolclass
        # logger.info(kwargs, prefix = 'Engine KWargs')
        return kwargs
    
    def get_session_kwargs(self, readonly: Optional[bool] = False, **session_kwargs) -> Dict[str, Any]:
        """
        Get the Session KWargs
        """
        kwargs = self.session_kwargs or {}
        if session_kwargs: kwargs = update_dict(kwargs, session_kwargs)
        kwargs = (
            update_dict(kwargs, self.session_ro_kwargs)
            if readonly
            else update_dict(kwargs, self.session_rw_kwargs)
        )
        # logger.info(kwargs, prefix = 'Session KWargs')
        return kwargs
    

    model_config = {'extra': 'allow', 'arbitrary_types_allowed': True}


class RemoteDatabaseConnection(PostgresConfig):
    """
    Remote Database
    """

    name: str
    extra_kws: Optional[Dict[str, Any]] = Field(default_factory = dict)
    _engine_rw: Optional[AsyncEngine] = PrivateAttr(None)
    _engine_ro: Optional[AsyncEngine] = PrivateAttr(None)
    _session_rw: Optional[async_sessionmaker[AsyncSession]] = PrivateAttr(None)
    _session_ro: Optional[async_sessionmaker[AsyncSession]] = PrivateAttr(None)

    @property
    def engine(self) -> AsyncEngine:
        """
        Returns the engine
        """
        if self._engine_rw is None:
            self._engine_rw = create_async_engine(**self.get_engine_kwargs(readonly = False, **self.extra_kws))
        return self._engine_rw
    
    @property
    def engine_ro(self) -> AsyncEngine:
        """
        Returns the readonly engine
        """
        if self._engine_ro is None:
            self._engine_ro = create_async_engine(**self.get_engine_kwargs(readonly = True, **self.extra_kws))
        return self._engine_ro
    
    @property
    def session_rw(self) -> async_sessionmaker[AsyncSession]:
        """
        Returns the session
        """
        if self._session_rw is None:
            self._session_rw = async_sessionmaker(
                self.engine, 
                class_ = AsyncSession, 
                **self.get_session_kwargs(readonly = False, **self.extra_kws)
            )
        return self._session_rw

    @property
    def session_ro(self) -> async_sessionmaker[AsyncSession]:
        """
        Returns the readonly session
        """
        if self._session_ro is None:
            self._session_ro = async_sessionmaker(
                self.engine_ro, class_ = AsyncSession, **self.get_session_kwargs(readonly = True, **self.extra_kws)
            )
        return self._session_ro



class DatabaseClientBase(abc.ABC):

    name: Optional[str] = 'database'
    

    def __init__(
        self, 
        settings: Optional[BaseSettings] = None,
        config: Optional[PostgresConfig] = None,
        **kwargs
    ):
        """
        Initializes the Database Client
        """
        # assert settings or config, 'Either settings or config must be provided'
        self.config: Optional[PostgresConfig] = None
        self._engine_rw: Optional[AsyncEngine] = None
        self._engine_ro: Optional[AsyncEngine] = None
        self._session_rw: Optional[async_sessionmaker[AsyncSession]] = None
        self._session_ro: Optional[async_sessionmaker[AsyncSession]] = None

        self._on_failure_callbacks: List[Callable] = []
        self._settings: Optional['AppSettings'] = None
        self._tpool: Optional['ThreadPool'] = None
        self._sql_template: Optional['SQLTemplates'] = None
        self._sql_template_kwargs: Optional[Dict[str, Any]] = {}

        # Remote Configs to Connect
        self.remote_connections: Dict[str, RemoteDatabaseConnection] = {}

        self.pre_init(**kwargs)
        if settings: self.config = PostgresConfig.from_settings(settings)
        elif config: self.config = config
        self.post_init(**kwargs)
        self._kwargs = kwargs
    

    def configure(
        self, 
        settings: Optional['AppSettings'] = None, 
        config: Optional[PostgresConfig] = None, 

        # Template Config
        template_name: Optional[str] = None,
        template_base_path: Optional[Union[str, Path]] = None,
        template_file_suffix: Optional[str] = None,
        **kwargs
    ):
        """
        Configures the Database Client
        """
        if settings:
            if hasattr(settings, 'ctx'):
                self._settings = settings
            if hasattr(settings, 'db'):
                if hasattr(settings.db, 'postgres'):
                    self.config = PostgresConfig.from_settings(settings.db.postgres)
                elif hasattr(settings.db, 'url'):
                    self.config = PostgresConfig.from_settings(settings.db)
            elif hasattr(settings, 'postgres'):
                self.config = PostgresConfig.from_settings(settings.postgres)
            
        # if settings: self.config = PostgresConfig.from_settings(settings)
        elif config: self.config = config

        if template_name: self._sql_template_kwargs['name'] = template_name
        if template_base_path: self._sql_template_kwargs['base_path'] = template_base_path
        if template_file_suffix: self._sql_template_kwargs['file_suffix'] = template_file_suffix
        if kwargs: self._kwargs = update_dict(self._kwargs, kwargs)

    def configure_remote(
        self,
        url: str,
        name: str,
        readonly_url: Optional[str] = None,
        superuser_url: Optional[str] = None,
        overwrite: Optional[bool] = None,
        **kwargs
    ):
        """
        Configures the remote database
        """
        if name in self.remote_connections and not overwrite: raise ValueError(f'Remote Connection {name} already exists')
        _config = {
            'name': name,
            'url': url,
            'readonly_url': readonly_url,
            'superuser_url': superuser_url,
        }
        _config = {k:v for k, v in _config.items() if v is not None}
        _config['extra_kws'] = kwargs
        self.remote_connections[name] = RemoteDatabaseConnection.model_validate(_config)



    def pre_init(self, **kwargs):
        """
        Pre-Init
        """
        pass

    def post_init(self, **kwargs):
        """
        Post-Init
        """
        pass

    @property
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        return self.settings.logger
            
    @property
    def autologger(self) -> 'Logger':
        """
        Returns the autologger
        """
        return self.settings.autologger

    @property
    def tpool(self) -> 'ThreadPool':
        """
        Returns the ThreadPool
        """
        if not self._tpool:
            from lazyops.libs.pooler import ThreadPooler
            self._tpool = ThreadPooler
        return self._tpool
    
    def configure_sql_templates(self, **kwargs):
        """
        Configures the SQL Templates
        """
        pass


    @property
    def sql_template(self) -> 'SQLTemplates':
        """
        Returns the SQL Templates
        """
        if not self._sql_template:
            from .templates import SQLTemplates
            self.configure_sql_templates()
            self._sql_template = SQLTemplates(
                settings = self.settings,
                **self._sql_template_kwargs
            )
        return self._sql_template

    @property
    def settings(self) -> 'AppSettings':
        """
        Returns the settings
        """
        # assert self._settings, 'The settings must be set'
        if self._settings is None:
            from lazyops.libs.abcs.configs.lazy import get_module_settings
            self._settings = get_module_settings(self.__module__)
        return self._settings

    def is_verbose_for(self, method: str) -> bool:
        """
        Returns if the method is verbose
        """
        if self.settings:
            return not self.settings.ctx.temp_data.has_logged(method)
        return True
    
    @property
    def engine(self) -> AsyncEngine:
        """
        Returns the engine
        """
        if self._engine_rw is None:
            assert self.config, 'The config must be set'
            self._engine_rw = create_async_engine(
                **self.config.get_engine_kwargs(
                    readonly = False, 
                    verbose = self.is_verbose_for('postgres_url'),
                    **self._kwargs
                )
            )
        return self._engine_rw
    
    @property
    def engine_ro(self) -> AsyncEngine:
        """
        Returns the readonly engine
        """
        if self._engine_ro is None:
            assert self.config, 'The config must be set'
            self._engine_ro = create_async_engine(
                **self.config.get_engine_kwargs(
                    readonly = True, 
                    verbose = self.is_verbose_for('postgres_readonly_warning'),
                    **self._kwargs
                )
            )
        return self._engine_ro
    
    @property
    def session_rw(self) -> async_sessionmaker[AsyncSession]:
        """
        Returns the session
        """
        if self._session_rw is None:
            assert self.config, 'The config must be set'
            self._session_rw = async_sessionmaker(
                self.engine, 
                class_ = AsyncSession, 
                **self.config.get_session_kwargs(
                    readonly = False, 
                    # verbose = self.is_verbose_for('postgres_url'),
                    **self._kwargs
                )
            )
        return self._session_rw

    @property
    def session_ro(self) -> async_sessionmaker[AsyncSession]:
        """
        Returns the readonly session
        """
        if self._session_ro is None:
            assert self.config, 'The config must be set'
            self._session_ro = async_sessionmaker(
                self.engine_ro, class_ = AsyncSession, **self.config.get_session_kwargs(
                    readonly = True, 
                    # verbose = self.is_verbose_for('postgres_readonly_warning'),
                    **self._kwargs
                )
            )
        return self._session_ro
    
    def run_error_callbacks(self, e: Exception):
        """
        Runs the error callbacks
        """
        import traceback
        tb = traceback.format_exc()
        for cb in self._on_failure_callbacks:
            self.tpool.create_background_task(cb, error = e, message = tb)
    

    async def get_db(self, readonly: Optional[bool] = None) -> AsyncGenerator[AsyncSession, None]:
        """
        Returns the database session
        """
        sess_type = self.session_ro if readonly else self.session_rw
        try:
            sess = sess_type()
            yield sess
        except Exception as e:
            logger.trace('Session error', e, depth=2)
            self.run_error_callbacks(e)
        finally:
            await sess.close()


    def get_database(
        self, 
        readonly: Optional[bool] = None,
        auto_commit: Optional[bool] = None,
        raise_errors: Optional[bool] = None,
    ) -> Callable[..., AsyncSession]:
        """
        Creates an inner dependency wrapper for the database session [FastAPI]
        """
        async def inner() -> AsyncGenerator[AsyncSession, None]:
            """
            Returns the database session
            """
            sess_type = self.session_ro if readonly else self.session_rw
            try:
                sess = sess_type()
                yield sess
            except Exception as e:
                logger.trace('Session error', e, depth=2)
                self.run_error_callbacks(e)
                if raise_errors: raise e
            finally:
                if auto_commit: await sess.commit()
                await sess.close()

        return inner
    

    @contextlib.asynccontextmanager
    async def session(
        self, 
        readonly: Optional[bool] = None,
        auto_commit: Optional[bool] = None,
        raise_errors: Optional[bool] = None,
        auto_rollback: Optional[bool] = True,
    ) -> AsyncGenerator[AsyncSession, None]:
        """
        Context manager for the database session
        """
        sess_type = self.session_ro if readonly else self.session_rw
        sess: AsyncSession = None
        try:
            sess: AsyncSession = sess_type()
            yield sess
        except Exception as e:
            logger.trace('Session error', e, depth = 5)
            self.run_error_callbacks(e)
            if auto_rollback: 
                try:
                    await sess.rollback()
                    logger.info('Rolled back session')
                except Exception as e:
                    logger.trace('Rollback error', e)
            if raise_errors: raise e
        finally:
            if auto_commit: await sess.commit()
            await sess.close()
            if readonly: gc.collect()

    @contextlib.asynccontextmanager
    async def session_readonly(
        self,
        raise_errors: Optional[bool] = None,
    ) -> AsyncGenerator[AsyncSession, None]:
        """
        Context manager for the database session
        """
        try:
            sess = self.session_ro()
            yield sess
        except Exception as e:
            logger.trace('Session error', e, depth = 2)
            self.run_error_callbacks(e)
            if raise_errors: raise e
        finally:
            await sess.close()
            gc.collect()


    @contextlib.asynccontextmanager
    async def remote_session(
        self, 
        name: str,
        readonly: Optional[bool] = None,
        auto_commit: Optional[bool] = None,
        raise_errors: Optional[bool] = None,
        auto_rollback: Optional[bool] = True,
    ) -> AsyncGenerator[AsyncSession, None]:
        """
        Context manager for the remote database session
        """
        if name not in self.remote_connections: raise ValueError(f'Remote Connection {name} does not exist')
        remote_conn = self.remote_connections[name]
        sess_type = remote_conn.session_ro if readonly else remote_conn.session_rw
        sess: AsyncSession = None
        try:
            sess: AsyncSession = sess_type()
            yield sess
        except Exception as e:
            logger.trace('Session error', e, depth = 5)
            self.run_error_callbacks(e)
            if auto_rollback: 
                try:
                    await sess.rollback()
                    logger.info('Rolled back session')
                except Exception as e:
                    logger.trace('Rollback error', e)
            if raise_errors: raise e
        finally:
            if auto_commit: await sess.commit()
            await sess.close()
            if readonly: gc.collect()



    @staticmethod
    def text(sql: str) -> TextClause:
        """
        Returns the text
        """
        return sql_text(sql)
    
    """
    Utility Methods
    """

    async def wait_for_ready(
        self,
        max_attempts: Optional[int] = 10,
        interval: Optional[float] = 5.0,
    ):
        """
        Waits for the database to be ready
        """
        attempts = 0
        while attempts < max_attempts:
            try:
                async with self.engine.begin() as conn:
                    await conn.execute(self.text('SELECT 1'))
                return
            except Exception as e:
                logger.info(f'Database not ready: {attempts}/{max_attempts} attempts. Error: {e}')
                attempts += 1
                await asyncio.sleep(interval)
                if attempts >= max_attempts:
                    raise e
        

    async def cast_result_to_object(
        self,
        result: 'Result',
        obj: 'DeclarativeBase',
        fields: Optional[List[str]],
        **kwargs,
    ) -> List['ModelType']:
        """
        Casts the result to the object
        """
        fields = fields or obj.__table__.columns.keys()
        return [obj(**dict(zip(fields, i))) for i in result.all()]
    
    async def bulk_commit_and_refresh(
        self,
        db: AsyncSession, 
        orm_objects: List['ModelType'],
        **kwargs,
    ) -> List['ModelType']:
        """
        Runs a bulk commit and refresh
        """
        await db.commit()
        for orm_object in orm_objects:
            await db.refresh(orm_object)
        return orm_objects
    

    async def upgrade_extensions(
        self,
        extensions: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        Upgrade the extensions
        """
        if not extensions: return
        async with self.engine.begin() as conn:
            for ext in extensions:
                await conn.execute(self.text(f"CREATE EXTENSION IF NOT EXISTS {ext}"))
                await conn.commit()
                await conn.execute(self.text(f"ALTER EXTENSION {ext} UPDATE"))
                await conn.commit()

                # Get the version
                result = await conn.execute(self.text(f"SELECT * FROM pg_extension WHERE extname = '{ext}'"))
                version = await result.scalar()
                logger.info(f"Extension {ext} version: {version}")

    
    def build_sql_filter_query(
        self,
        **filters: Dict[str, Union[int, float, datetime.datetime, Dict, List, Any]]
    ) -> List[str]: 
        """
        Constructs the WHERE clause of a SQL query

        - This is a generic method that can be used to build the WHERE clause of a SQL query
        """
        raise NotImplementedError
    

    def build_sql_filter(
        self,
        conditional: Optional[str] = 'AND',
        **filters: Dict[str, Union[int, float, datetime.datetime, Dict, List, Any]]
    ) -> Dict[str, Union[List[str], str]]:
        """
        Returns 
        {
            "conditional": "AND",
            "statements": [
                "statement1",
                "statement2",
                "statement3",
            ]
        }
        which is properly formatted for the template
        """
        return {
            "conditional": conditional,
            "statements": self.build_sql_filter_query(**filters),
        }
    
    

    def build_sql_filters(
        self,
        and_filters: Optional[Dict[str, Union[int, float, datetime.datetime, Dict, List, Any]]] = None,
        or_filters: Optional[Dict[str, Union[int, float, datetime.datetime, Dict, List, Any]]] = None,
    ) -> List[Dict[str, Union[List[str], str]]]:
        """
        Creates the proper SQL filters
        [
            {
                "conditional": "AND",
                "statements": [
                    "statement1",
                    "statement2",
                    "statement3",
                ]
            },
            {
                "conditional": "OR",
                "statements": [
                    "statement4",
                    "statement5",
                    "statement6",
                ]
            }
        ]
        which is properly formatted for the template
        """
        filters = []
        if and_filters:
            filters.append(self.build_sql_filter(conditional = "AND", **and_filters))
        if or_filters:
            filters.append(self.build_sql_filter(conditional = 'OR', **or_filters))
        return filters
    

    async def execute_sql(
        self,
        *ops: str,
        db: Optional[AsyncSession] = None,
        readonly: Optional[bool] = None,
        auto_commit: Optional[bool] = None,
        raise_errors: Optional[bool] = None,
        verbose: Optional[bool] = None,
        filters: Optional[Dict[str, Union[int, float, datetime.datetime, Dict, List, Any]]] = None,
        **kwargs: Any,
    ) -> 'Result':
        """
        Executes a SQL query from a template
        """

        filters = self.build_sql_filters(and_filters=filters) if filters else None
        sql_query = await self.sql_template.arender(*ops, filters = filters, **kwargs)
        if verbose: logger.info(sql_query, prefix = 'Executing SQL Query')
        if db is not None: return await db.execute(self.text(sql_query))
        async with self.session(readonly = readonly, auto_commit = auto_commit, raise_errors = raise_errors) as db:
            return await db.execute(self.text(sql_query))

    
    async def index_exists(
        self,
        index_name: str,
    ) -> bool:
        """
        Checks if the Index Exists
        """
        statement = self.text(
            """
            SELECT EXISTS (
                SELECT FROM pg_indexes
                WHERE indexname = :index_name
            );
            """
        )
        async with self.session() as session:
            result = await session.execute(statement, {'index_name': index_name})
            return result.scalar_one()


    async def database_exists(
        self,
        db_name: str,
    ) -> bool:
        """
        Checks if the database exists
        """
        statement = self.text(
            f"SELECT 1 FROM pg_database WHERE datname = '{db_name}';"
        )
        async with self.session() as session:
            result = await session.execute(statement)
            output = result.scalar_one()
            logger.info(f'Database Exists: {output}')
            return output
        


import os
import signal
from typing import Optional, List, TYPE_CHECKING

if TYPE_CHECKING:
    from lazyops.utils.logs import Logger


class GracefulKiller:
    kill_now = False
    signals = {
        signal.SIGINT: 'SIGINT',
        signal.SIGTERM: 'SIGTERM',
        signal.SIGKILL: 'SIGKILL',
        signal.SIGQUIT: 'SIGQUIT',
        signal.SIGSTOP: 'SIGSTOP',
        signal.SIGABRT: 'SIGABRT',
    }

    def __init__(
        self, 
        enabled: Optional[List[signal.Signals]] = [signal.SIGINT, signal.SIGTERM],
        logger: Optional['Logger'] = None
    ):  # sourcery skip: default-mutable-arg
        
        self.enabled = enabled
        if logger is None:
            from lazyops.utils.logs import logger as _logger
            logger = _logger
        self.logger = logger
        for sig in self.enabled:
            signal.signal(sig, self.exit_gracefully)
        self.pid = os.getpid()

    def exit_gracefully(self, signum, frame):
        self.logger.warning(f"[{self.pid}] Received {self.signals[signum]} signal. Exiting gracefully")
        self.kill_now = True


from __future__ import annotations

"""
Extra Network Types
"""


import contextlib
import functools
import urllib.request
from typing import Union, List, Optional
from typing_extensions import Annotated

from pydantic import (
    BeforeValidator,
    AfterValidator,
    PlainSerializer,
    TypeAdapter,
    WithJsonSchema,
)


@functools.lru_cache(maxsize = 128)
def validate_one_url(url: str) -> bool:
    """
    Validate that the url is valid
    """
    with contextlib.suppress(Exception):
        urllib.request.urlopen(url, timeout = 0.5)
        return True
    return False


def validate_url_from_list(urls: Union[str, List[str]]) -> Optional[str]:
    """
    Validates between the urls and returns the first valid url
    """
    urls = [urls] if isinstance(urls, str) else urls
    return next((url for url in urls if validate_one_url(url)), None)

"""
Usage:

class ServerSettings(BaseModel):
    endpoint: ValidURL

settings = ServerSettings(endpoint = "https://www.google.com")
settings = ServerSettings(endpoint = ["http://fake.local", "https://www.google.com"])

>> settings.endpoint = "https://www.google.com"
"""

ValidURL = Annotated[
    str,
    BeforeValidator(validate_url_from_list),
    WithJsonSchema({'type': 'string'}, mode='serialization'),
]


